"""
Plain DETR, adapted from
https://github.com/impiga/Plain-DETR

Paper "DETR Doesn't Need Multi-Scale or Locality Design", https://arxiv.org/abs/2308.01904

Changes from original:
* Move background index to first from last (to be inline with the rest of Birder detectors)
* Removed two stage support
* Only support pre-norm (original supports both pre- and post-norm)
"""

# Reference license: MIT

import copy
import math
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import boxes as box_ops
from torchvision.ops import sigmoid_focal_loss

from birder.common import training_utils
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.deformable_detr import HungarianMatcher
from birder.net.detection.deformable_detr import inverse_sigmoid
from birder.net.detection.detr import PositionEmbeddingSine
from birder.ops.soft_nms import SoftNMS


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class MultiheadAttention(nn.Module):
    def __init__(self, d_model: int, num_heads: int, attn_drop: float = 0.0, proj_drop: float = 0.0) -> None:
        super().__init__()
        assert d_model % num_heads == 0, "d_model should be divisible by num_heads"

        self.num_heads = num_heads
        self.head_dim = d_model // num_heads
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.v_proj = nn.Linear(d_model, d_model)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(d_model, d_model)
        self.proj_drop = nn.Dropout(proj_drop)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.proj.weight)
        if self.q_proj.bias is not None:
            nn.init.zeros_(self.q_proj.bias)
            nn.init.zeros_(self.k_proj.bias)
            nn.init.zeros_(self.v_proj.bias)
            nn.init.zeros_(self.proj.bias)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        B, l_q, C = query.size()
        q = self.q_proj(query).reshape(B, l_q, self.num_heads, self.head_dim).transpose(1, 2)
        k = self.k_proj(key).reshape(B, key.size(1), self.num_heads, self.head_dim).transpose(1, 2)
        v = self.v_proj(value).reshape(B, value.size(1), self.num_heads, self.head_dim).transpose(1, 2)

        if key_padding_mask is not None:
            # key_padding_mask is expected to be boolean (True = masked)
            # SDPA expects True = attend, so we invert
            attn_mask = ~key_padding_mask[:, None, None, :]
        else:
            attn_mask = None

        attn = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=attn_mask, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        attn = attn.transpose(1, 2).reshape(B, l_q, C)
        x = self.proj(attn)
        x = self.proj_drop(x)

        return x


class GlobalCrossAttention(nn.Module):
    """
    Global cross-attention with Box-to-Pixel Relative Position Bias (BoxRPB)

    This utilizes Box-to-Pixel Relative Position Bias (BoxRPB) to guide attention
    using the spatial relationship between query boxes and image features.
    The bias calculation is decomposed into axial (x and y) components.
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout: float,
        rpe_hidden_dim: int,
        feature_stride: int,
        rpe_type: Literal["linear", "log"],
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.scale = self.head_dim**-0.5
        self.feature_stride = feature_stride
        self.rpe_type = rpe_type

        self.q_proj = nn.Linear(embed_dim, embed_dim)
        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.out_proj = nn.Linear(embed_dim, embed_dim)

        self.attn_drop = nn.Dropout(dropout)
        self.proj_drop = nn.Dropout(dropout)

        self.cpb_mlp_x = nn.Sequential(
            nn.Linear(2, rpe_hidden_dim),
            nn.ReLU(inplace=True),
            nn.Linear(rpe_hidden_dim, num_heads, bias=False),
        )
        self.cpb_mlp_y = nn.Sequential(
            nn.Linear(2, rpe_hidden_dim),
            nn.ReLU(inplace=True),
            nn.Linear(rpe_hidden_dim, num_heads, bias=False),
        )

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        reference_points: torch.Tensor,
        spatial_shape: tuple[int, int],
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        B, num_queries, _ = query.size()
        H, W = spatial_shape

        q = self.q_proj(query)
        k = self.k_proj(key)
        v = self.v_proj(value)

        q = q.view(B, num_queries, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        k = k.view(B, H * W, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        v = v.view(B, H * W, self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        q = q * self.scale

        attn = q @ k.transpose(-2, -1)
        rpe = self._compute_box_rpe(reference_points, H, W, query.device)
        attn = attn + rpe

        if key_padding_mask is not None:
            attn = attn.masked_fill(key_padding_mask[:, None, None, :], float("-inf"))

        attn = F.softmax(attn, dim=-1)
        attn = self.attn_drop(attn)

        out = attn @ v
        out = out.permute(0, 2, 1, 3).reshape(B, num_queries, self.embed_dim)
        out = self.out_proj(out)
        out = self.proj_drop(out)

        return out

    # pylint: disable=too-many-locals
    def _compute_box_rpe(self, reference_points: torch.Tensor, H: int, W: int, device: torch.device) -> torch.Tensor:
        B, n_q, _ = reference_points.size()
        stride = self.feature_stride

        # cxcywh to xyxy
        cx, cy, bw, bh = reference_points.unbind(-1)
        x1 = cx - bw / 2
        y1 = cy - bh / 2
        x2 = cx + bw / 2
        y2 = cy + bh / 2

        # Scale to pixel coordinates
        x1 = x1 * (W * stride)
        y1 = y1 * (H * stride)
        x2 = x2 * (W * stride)
        y2 = y2 * (H * stride)

        # Pixel grid (cell centers)
        pos_x = torch.linspace(0.5, W - 0.5, W, device=device) * stride
        pos_y = torch.linspace(0.5, H - 0.5, H, device=device) * stride

        # Box edge to pixel distances
        delta_x1 = x1[:, :, None] - pos_x[None, None, :]
        delta_x2 = x2[:, :, None] - pos_x[None, None, :]
        delta_y1 = y1[:, :, None] - pos_y[None, None, :]
        delta_y2 = y2[:, :, None] - pos_y[None, None, :]

        if self.rpe_type == "log":
            delta_x1 = torch.sign(delta_x1) * torch.log2(torch.abs(delta_x1) + 1.0) / 3.0
            delta_x2 = torch.sign(delta_x2) * torch.log2(torch.abs(delta_x2) + 1.0) / 3.0
            delta_y1 = torch.sign(delta_y1) * torch.log2(torch.abs(delta_y1) + 1.0) / 3.0
            delta_y2 = torch.sign(delta_y2) * torch.log2(torch.abs(delta_y2) + 1.0) / 3.0

        delta_x = torch.stack([delta_x1, delta_x2], dim=-1)
        delta_y = torch.stack([delta_y1, delta_y2], dim=-1)

        rpe_x = self.cpb_mlp_x(delta_x)
        rpe_y = self.cpb_mlp_y(delta_y)

        # Axial decomposition: rpe[h,w] = rpe_y[h] + rpe_x[w]
        rpe = rpe_y[:, :, :, None, :] + rpe_x[:, :, None, :, :]
        rpe = rpe.reshape(B, n_q, H * W, self.num_heads)
        rpe = rpe.permute(0, 3, 1, 2)

        return rpe


class GlobalDecoderLayer(nn.Module):
    """
    Transformer decoder layer with global cross-attention and BoxRPB
    """

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_feedforward: int,
        dropout: float,
        rpe_hidden_dim: int,
        feature_stride: int,
        rpe_type: Literal["linear", "log"],
    ) -> None:
        super().__init__()

        self.self_attn = MultiheadAttention(d_model, num_heads, attn_drop=dropout)
        self.cross_attn = GlobalCrossAttention(
            embed_dim=d_model,
            num_heads=num_heads,
            dropout=dropout,
            rpe_hidden_dim=rpe_hidden_dim,
            feature_stride=feature_stride,
            rpe_type=rpe_type,
        )

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.activation = nn.ReLU()

    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        query_pos: torch.Tensor,
        memory_pos: torch.Tensor,
        reference_points: torch.Tensor,
        spatial_shape: tuple[int, int],
        memory_key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        tgt2 = self.norm1(tgt)
        qk = tgt2 + query_pos
        tgt2 = self.self_attn(qk, qk, tgt2)
        tgt = tgt + self.dropout1(tgt2)

        tgt2 = self.cross_attn(
            query=self.norm2(tgt) + query_pos,
            key=memory + memory_pos,
            value=memory,
            reference_points=reference_points,
            spatial_shape=spatial_shape,
            key_padding_mask=memory_key_padding_mask,
        )
        tgt = tgt + self.dropout2(tgt2)

        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(self.norm3(tgt)))))
        tgt = tgt + self.dropout3(tgt2)

        return tgt


class GlobalDecoder(nn.Module):
    def __init__(self, decoder_layer: nn.Module, num_layers: int, norm: nn.Module, d_model: int) -> None:
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        self.d_model = d_model

        self.bbox_embed: Optional[nn.ModuleList] = None
        self.class_embed: Optional[nn.ModuleList] = None

        self.reset_parameters()

    def reset_parameters(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.LayerNorm):
                nn.init.zeros_(m.bias)
                nn.init.ones_(m.weight)

        for m in self.modules():
            if m is not self and hasattr(m, "reset_parameters") is True and callable(m.reset_parameters) is True:
                m.reset_parameters()

    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        query_pos: torch.Tensor,
        memory_pos: torch.Tensor,
        reference_points: torch.Tensor,
        spatial_shape: tuple[int, int],
        memory_key_padding_mask: Optional[torch.Tensor] = None,
        return_intermediates: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        output = tgt
        intermediate = []
        intermediate_reference_points = []

        if self.bbox_embed is not None:
            for layer, bbox_embed in zip(self.layers, self.bbox_embed):
                reference_points_input = reference_points.detach().clamp(0, 1)

                output = layer(
                    output,
                    memory,
                    query_pos=query_pos,
                    memory_pos=memory_pos,
                    reference_points=reference_points_input,
                    spatial_shape=spatial_shape,
                    memory_key_padding_mask=memory_key_padding_mask,
                )

                output_for_pred = self.norm(output)
                tmp = bbox_embed(output_for_pred)
                new_reference_points = tmp + inverse_sigmoid(reference_points)
                new_reference_points = new_reference_points.sigmoid()
                reference_points = new_reference_points.detach()

                if return_intermediates is True:
                    intermediate.append(output_for_pred)
                    intermediate_reference_points.append(new_reference_points)

            if return_intermediates is True:
                return torch.stack(intermediate), torch.stack(intermediate_reference_points)

            return output_for_pred, new_reference_points

        for layer in self.layers:
            reference_points_input = reference_points.detach().clamp(0, 1)

            output = layer(
                output,
                memory,
                query_pos=query_pos,
                memory_pos=memory_pos,
                reference_points=reference_points_input,
                spatial_shape=spatial_shape,
                memory_key_padding_mask=memory_key_padding_mask,
            )

            output_for_pred = self.norm(output)

            if return_intermediates is True:
                intermediate.append(output_for_pred)
                intermediate_reference_points.append(reference_points)

        if return_intermediates is True:
            return torch.stack(intermediate), torch.stack(intermediate_reference_points)

        return output_for_pred, reference_points


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model: int, num_heads: int, dim_feedforward: int, dropout: float) -> None:
        super().__init__()
        self.self_attn = MultiheadAttention(d_model, num_heads, attn_drop=dropout)

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = nn.ReLU()

    def forward(
        self, src: torch.Tensor, pos: torch.Tensor, src_key_padding_mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        src2 = self.norm1(src)
        q = src2 + pos
        k = src2 + pos

        src2 = self.self_attn(q, k, src2, key_padding_mask=src_key_padding_mask)
        src = src + self.dropout1(src2)

        src2 = self.linear2(self.dropout(self.activation(self.linear1(self.norm2(src)))))
        src = src + self.dropout2(src2)

        return src


class TransformerEncoder(nn.Module):
    def __init__(self, encoder_layer: nn.Module, num_layers: int) -> None:
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)

    def forward(self, x: torch.Tensor, pos: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        out = x
        for layer in self.layers:
            out = layer(out, pos=pos, src_key_padding_mask=mask)

        return out


# pylint: disable=invalid-name
class Plain_DETR(DetectionBaseNet):
    default_size = (640, 640)
    block_group_regex = r"encoder\.layers\.(\d+)|decoder\.layers\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        # Sigmoid based classification (like multi-label networks)
        self.num_classes = self.num_classes - 1

        hidden_dim = 256
        num_heads = 8
        dropout = 0.0
        dim_feedforward: int = self.config.get("dim_feedforward", 2048)
        num_encoder_layers: int = self.config["num_encoder_layers"]
        num_decoder_layers: int = self.config["num_decoder_layers"]
        num_queries_one2one: int = self.config.get("num_queries_one2one", 300)
        num_queries_one2many: int = self.config.get("num_queries_one2many", 0)
        k_one2many: int = self.config.get("k_one2many", 6)
        lambda_one2many: float = self.config.get("lambda_one2many", 1.0)
        rpe_hidden_dim: int = self.config.get("rpe_hidden_dim", 512)
        rpe_type: Literal["linear", "log"] = self.config.get("rpe_type", "linear")
        box_refine: bool = self.config.get("box_refine", True)
        soft_nms: bool = self.config.get("soft_nms", False)

        self.soft_nms = None
        if soft_nms is True:
            self.soft_nms = SoftNMS()

        self.hidden_dim = hidden_dim
        self.num_queries_one2one = num_queries_one2one
        self.num_queries_one2many = num_queries_one2many
        self.k_one2many = k_one2many
        self.lambda_one2many = lambda_one2many
        self.box_refine = box_refine
        self.num_queries = self.num_queries_one2one + self.num_queries_one2many
        if hasattr(self.backbone, "max_stride") is True:
            self.feature_stride = self.backbone.max_stride
        else:
            self.feature_stride = 32

        if num_encoder_layers == 0:
            self.encoder = None
        else:
            encoder_layer = TransformerEncoderLayer(hidden_dim, num_heads, dim_feedforward, dropout)
            self.encoder = TransformerEncoder(encoder_layer, num_encoder_layers)

        decoder_layer = GlobalDecoderLayer(
            hidden_dim,
            num_heads=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            rpe_hidden_dim=rpe_hidden_dim,
            feature_stride=self.feature_stride,
            rpe_type=rpe_type,
        )
        decoder_norm = nn.LayerNorm(hidden_dim)
        self.decoder = GlobalDecoder(
            decoder_layer,
            num_decoder_layers,
            decoder_norm,
            d_model=hidden_dim,
        )

        self.class_embed = nn.Linear(hidden_dim, self.num_classes)
        self.bbox_embed = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)
        self.query_embed = nn.Parameter(torch.empty(self.num_queries, hidden_dim * 2))
        self.reference_point_head = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)
        self.input_proj = nn.Conv2d(
            self.backbone.return_channels[-1], hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.pos_enc = PositionEmbeddingSine(hidden_dim // 2, normalize=True)
        self.matcher = HungarianMatcher(cost_class=2.0, cost_bbox=5.0, cost_giou=2.0)

        if box_refine is True:
            self.class_embed = _get_clones(self.class_embed, num_decoder_layers)
            self.bbox_embed = _get_clones(self.bbox_embed, num_decoder_layers)
            self.decoder.bbox_embed = self.bbox_embed
        else:
            self.class_embed = nn.ModuleList([self.class_embed for _ in range(num_decoder_layers)])
            self.bbox_embed = nn.ModuleList([self.bbox_embed for _ in range(num_decoder_layers)])

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

        # Weights initialization
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        for class_embed in self.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

        for idx, bbox_embed in enumerate(self.bbox_embed):
            last_linear = [m for m in bbox_embed.modules() if isinstance(m, nn.Linear)][-1]
            nn.init.zeros_(last_linear.weight)
            nn.init.zeros_(last_linear.bias)
            if idx == 0:
                nn.init.constant_(last_linear.bias[2:], -2.0)  # Small initial wh

        nn.init.normal_(self.query_embed)
        ref_last_linear = [m for m in self.reference_point_head.modules() if isinstance(m, nn.Linear)][-1]
        nn.init.zeros_(ref_last_linear.weight)
        nn.init.zeros_(ref_last_linear.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        num_decoder_layers = len(self.class_embed)
        self.class_embed = nn.ModuleList([nn.Linear(self.hidden_dim, num_classes) for _ in range(num_decoder_layers)])

        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        for class_embed in self.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad_(False)

        if freeze_classifier is False:
            for param in self.class_embed.parameters():
                param.requires_grad_(True)

    @staticmethod
    def _get_src_permutation_idx(indices: list[tuple[torch.Tensor, torch.Tensor]]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _class_loss(
        self,
        cls_logits: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: int,
    ) -> torch.Tensor:
        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)

        target_classes_onehot = torch.zeros(
            cls_logits.size(0),
            cls_logits.size(1),
            cls_logits.size(2) + 1,
            dtype=cls_logits.dtype,
            device=cls_logits.device,
        )
        target_classes_onehot[idx[0], idx[1], target_classes_o] = 1
        target_classes_onehot = target_classes_onehot[:, :, :-1]

        loss = sigmoid_focal_loss(cls_logits, target_classes_onehot, alpha=0.25, gamma=2.0)
        loss_ce = (loss.mean(1).sum() / num_boxes) * cls_logits.size(1)

        return loss_ce

    def _box_loss(
        self,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: int,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)
        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none")
        loss_bbox = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(
            box_ops.generalized_box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        )
        loss_giou = loss_giou.sum() / num_boxes

        return (loss_bbox, loss_giou)

    @torch.jit.unused  # type: ignore[untyped-decorator]
    @torch.compiler.disable()  # type: ignore[untyped-decorator]
    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
        cls_logits_one2many: Optional[torch.Tensor] = None,
        box_output_one2many: Optional[torch.Tensor] = None,
    ) -> dict[str, torch.Tensor]:
        # Compute the average number of target boxes across all nodes, for normalization purposes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=cls_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1)

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []
        for idx in range(cls_logits.size(0)):
            indices = self.matcher(cls_logits[idx], box_output[idx], targets)
            loss_ce_i = self._class_loss(cls_logits[idx], targets, indices, num_boxes)
            loss_bbox_i, loss_giou_i = self._box_loss(box_output[idx], targets, indices, num_boxes)
            loss_ce_list.append(loss_ce_i)
            loss_bbox_list.append(loss_bbox_i)
            loss_giou_list.append(loss_giou_i)

        loss_ce = torch.stack(loss_ce_list).sum() * 2
        loss_bbox = torch.stack(loss_bbox_list).sum() * 5
        loss_giou = torch.stack(loss_giou_list).sum() * 2

        # One2many loss (hybrid matching)
        if cls_logits_one2many is not None and box_output_one2many is not None:
            targets_one2many = [
                {"boxes": t["boxes"].repeat(self.k_one2many, 1), "labels": t["labels"].repeat(self.k_one2many)}
                for t in targets
            ]
            num_boxes_one2many = num_boxes * self.k_one2many

            loss_ce_list_one2many = []
            loss_bbox_list_one2many = []
            loss_giou_list_one2many = []
            for idx in range(cls_logits_one2many.size(0)):
                indices = self.matcher(cls_logits_one2many[idx], box_output_one2many[idx], targets_one2many)
                loss_ce_i = self._class_loss(cls_logits_one2many[idx], targets_one2many, indices, num_boxes_one2many)
                loss_bbox_i, loss_giou_i = self._box_loss(
                    box_output_one2many[idx], targets_one2many, indices, num_boxes_one2many
                )
                loss_ce_list_one2many.append(loss_ce_i)
                loss_bbox_list_one2many.append(loss_bbox_i)
                loss_giou_list_one2many.append(loss_giou_i)

            loss_ce += torch.stack(loss_ce_list_one2many).sum() * 2 * self.lambda_one2many
            loss_bbox += torch.stack(loss_bbox_list_one2many).sum() * 5 * self.lambda_one2many
            loss_giou += torch.stack(loss_giou_list_one2many).sum() * 2 * self.lambda_one2many

        losses = {
            "labels": loss_ce,
            "boxes": loss_bbox,
            "giou": loss_giou,
        }

        return losses

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_sizes: torch.Tensor
    ) -> list[dict[str, torch.Tensor]]:
        prob = class_logits.sigmoid()
        scores, labels = prob.max(-1)
        labels = labels + 1  # Background offset

        # Convert to [x0, y0, x1, y1] format
        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")

        # Convert from relative [0, 1] to absolute [0, height] coordinates
        img_h, img_w = image_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, l, b in zip(scores, labels, boxes):
            # Non-maximum suppression
            if self.soft_nms is not None:
                soft_scores, keep = self.soft_nms(b, s, l, score_threshold=0.001)
                s[keep] = soft_scores

                b = b[keep]
                s = s[keep]
                l = l[keep]  # noqa: E741

            detections.append(
                {
                    "boxes": b,
                    "scores": s,
                    "labels": l,
                }
            )

        return detections

    def forward_net(self, x: torch.Tensor, masks: Optional[torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        src = features[self.backbone.return_stages[-1]]
        src = self.input_proj(src)
        B, _, H, W = src.size()

        if masks is not None:
            masks = F.interpolate(masks[None].float(), size=(H, W), mode="nearest").to(torch.bool)[0]
            mask_flatten = masks.flatten(1)
        else:
            mask_flatten = None

        pos = self.pos_enc(src, masks)
        src = src.flatten(2).permute(0, 2, 1)
        pos = pos.flatten(2).permute(0, 2, 1)

        if self.encoder is not None:
            memory = self.encoder(src, pos=pos, mask=mask_flatten)
        else:
            memory = src

        # Use all queries during training, only one2one during inference
        if self.training is True and self.num_queries_one2many > 0:
            num_queries_to_use = self.num_queries_one2one + self.num_queries_one2many
        else:
            num_queries_to_use = self.num_queries_one2one

        query_embed = self.query_embed[:num_queries_to_use]
        query_embed, query_pos = torch.split(query_embed, self.hidden_dim, dim=1)
        query_embed = query_embed.unsqueeze(0).expand(B, -1, -1)
        query_pos = query_pos.unsqueeze(0).expand(B, -1, -1)

        reference_points = self.reference_point_head(query_pos).sigmoid()

        hs, inter_references = self.decoder(
            tgt=query_embed,
            memory=memory,
            query_pos=query_pos,
            memory_pos=pos,
            reference_points=reference_points,
            spatial_shape=(H, W),
            memory_key_padding_mask=mask_flatten,
            return_intermediates=self.training is True,
        )

        if self.training is True:
            outputs_classes = []
            outputs_coords = []
            for lvl, (class_embed, bbox_embed) in enumerate(zip(self.class_embed, self.bbox_embed)):
                outputs_class = class_embed(hs[lvl])
                outputs_classes.append(outputs_class)

                if self.box_refine is True:
                    outputs_coord = inter_references[lvl]
                else:
                    tmp = bbox_embed(hs[lvl])
                    tmp = tmp + inverse_sigmoid(reference_points)
                    outputs_coord = tmp.sigmoid()

                outputs_coords.append(outputs_coord)

            outputs_class = torch.stack(outputs_classes)
            outputs_coord = torch.stack(outputs_coords)
        else:
            class_embed = self.class_embed[-1]
            bbox_embed = self.bbox_embed[-1]
            outputs_class = class_embed(hs)

            if self.box_refine is True:
                outputs_coord = inter_references
            else:
                tmp = bbox_embed(hs)
                tmp = tmp + inverse_sigmoid(reference_points)
                outputs_coord = tmp.sigmoid()

        return (outputs_class, outputs_coord)

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[tuple[int, int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        outputs_class, outputs_coord = self.forward_net(x, masks)

        losses = {}
        detections: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            assert targets is not None, "targets should not be none when in training mode"

            for idx, target in enumerate(targets):
                boxes = target["boxes"]
                boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
                scale = images.image_sizes[idx].flip(0).repeat(2).float()  # flip to [W, H], repeat to [W, H, W, H]
                boxes = boxes / scale
                targets[idx]["boxes"] = boxes
                targets[idx]["labels"] = target["labels"] - 1  # No background

            # Split outputs for one2one and one2many
            outputs_class_one2one = outputs_class[:, :, : self.num_queries_one2one]
            outputs_coord_one2one = outputs_coord[:, :, : self.num_queries_one2one]

            if self.num_queries_one2many > 0:
                outputs_class_one2many = outputs_class[:, :, self.num_queries_one2one :]
                outputs_coord_one2many = outputs_coord[:, :, self.num_queries_one2one :]
            else:
                outputs_class_one2many = None
                outputs_coord_one2many = None

            losses = self.compute_loss(
                targets, outputs_class_one2one, outputs_coord_one2one, outputs_class_one2many, outputs_coord_one2many
            )

        else:
            detections = self.postprocess_detections(outputs_class, outputs_coord, images.image_sizes)

        return (detections, losses)


registry.register_model_config(
    "plain_detr_lite",
    Plain_DETR,
    config={"num_encoder_layers": 1, "num_decoder_layers": 3, "box_refine": False},
)
registry.register_model_config(
    "plain_detr",
    Plain_DETR,
    config={"num_encoder_layers": 0, "num_decoder_layers": 6, "num_queries_one2many": 1500},
)



"""
FungiCLEF2023 benchmark using KNN for fungi species classification

Link: https://www.imageclef.org/FungiCLEF2023
"""

import argparse
import logging
import time
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import polars as pl
import torch
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.datahub.evaluation import FungiCLEF2023
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.knn import evaluate_knn
from birder.eval.methods.simpleshot import sample_k_shot

logger = logging.getLogger(__name__)


def _print_summary_table(results: list[dict[str, Any]], k_values: list[int]) -> None:
    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("FungiCLEF2023 (KNN)", style="dim")
    for k in k_values:
        table.add_column(f"k={k}", justify="right")

    table.add_column("Runs", justify="right")

    for result in results:
        row = [Path(result["embeddings_file"]).name]
        for k in k_values:
            acc = result["accuracies"].get(k)
            row.append(f"{acc:.4f}" if acc is not None else "-")

        row.append(f"{result['num_runs']}")
        table.add_row(*row)

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], k_values: list[int], output_path: Path) -> None:
    rows: list[dict[str, Any]] = []
    for result in results:
        row: dict[str, Any] = {
            "embeddings_file": result["embeddings_file"],
            "method": result["method"],
            "num_runs": result["num_runs"],
        }
        for k in k_values:
            row[f"k_{k}_acc"] = result["accuracies"].get(k)
            row[f"k_{k}_std"] = result["accuracies_std"].get(k)

        rows.append(row)

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _load_fungiclef_metadata(dataset: FungiCLEF2023) -> pl.DataFrame:
    """
    Load metadata from FungiCLEF2023 CSV files

    Returns DataFrame with columns: id (filename stem), label, split (train/val/test).
    Filters out validation samples with unknown species (class_id == settings.NO_LABEL).
    Test samples have label=settings.NO_LABEL (no ground truth available) and are excluded from evaluation.
    """

    train_df = pl.read_csv(dataset.train_metadata_path)
    train_df = train_df.with_columns(
        pl.col("image_path").map_elements(lambda p: Path(p).stem, return_dtype=pl.Utf8).alias("id"),
        pl.lit("train").alias("split"),
    ).select(["id", "class_id", "split"])

    val_df = pl.read_csv(dataset.val_metadata_path)
    val_df = val_df.filter(pl.col("class_id") != settings.NO_LABEL)
    val_df = val_df.with_columns(
        pl.col("filename").alias("id"),
        pl.lit("val").alias("split"),
    ).select(["id", "class_id", "split"])

    # Include test IDs so they are properly excluded when embeddings contain all samples
    test_df = pl.read_csv(dataset.test_metadata_path)
    test_df = test_df.with_columns(
        pl.col("filename").alias("id"),
        pl.lit("test").alias("split"),
        pl.lit(settings.NO_LABEL, dtype=pl.Int64).alias("class_id"),
    ).select(["id", "class_id", "split"])

    metadata_df = pl.concat([train_df, val_df, test_df])
    metadata_df = metadata_df.rename({"class_id": "label"})

    return metadata_df


def _load_embeddings_with_split(
    embeddings_path: str, metadata_df: pl.DataFrame
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.int_], npt.NDArray[np.float32], npt.NDArray[np.int_]]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    train_meta = metadata_df.filter(pl.col("split") == "train").select(["id", "label"])
    val_meta = metadata_df.filter(pl.col("split") == "val").select(["id", "label"])
    test_meta = metadata_df.filter(pl.col("split") == "test").select(["id"])

    train_join = train_meta.join(emb_df, on="id", how="inner")
    val_join = val_meta.join(emb_df, on="id", how="inner")
    test_join = test_meta.join(emb_df, on="id", how="inner")

    dropped_train = train_meta.height - train_join.height
    dropped_val = val_meta.height - val_join.height
    dropped_test = test_meta.height - test_join.height
    dropped_total = dropped_train + dropped_val + dropped_test
    if dropped_total > 0:
        logger.warning(
            f"Join dropped {dropped_total} samples (missing embeddings): "
            f"train={dropped_train}, val={dropped_val}, test={dropped_test}"
        )

    x_train = train_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_train = train_join.get_column("label").to_numpy().astype(np.int_)
    x_val = val_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_val = val_join.get_column("label").to_numpy().astype(np.int_)

    num_classes = y_train.max() + 1
    total_samples = x_train.shape[0] + x_val.shape[0] + test_join.height
    embedding_dim = x_train.shape[1]
    logger.info(f"Loaded {total_samples} samples with {embedding_dim} dimensions, {num_classes} classes")

    logger.info(
        f"Train: {len(y_train)} samples, Val: {len(y_val)} samples, "
        f"Test: {test_join.height} samples (no labels, excluded)"
    )

    return (x_train, y_train, x_val, y_val)


def _evaluate_single_k(
    x_train: npt.NDArray[np.float32],
    y_train: npt.NDArray[np.int_],
    x_val: npt.NDArray[np.float32],
    y_val: npt.NDArray[np.int_],
    k: int,
    num_runs: int,
    seed: int,
    device: torch.device,
) -> tuple[float, float]:
    logger.info(f"Evaluating k={k} ({k}-shot sampling, KNN k={k})")

    scores: list[float] = []
    for run in range(num_runs):
        run_seed = seed + run
        rng = np.random.default_rng(run_seed)

        # Sample k examples per class
        x_train_k, y_train_k = sample_k_shot(x_train, y_train, k, rng)

        # Evaluate using KNN with k neighbors
        y_pred, y_true = evaluate_knn(x_train_k, y_train_k, x_val, y_val, k=k, device=device)

        acc = float(np.mean(y_pred == y_true))
        scores.append(acc)
        logger.info(f"Run {run + 1}/{num_runs} - Accuracy: {acc:.4f}")

    scores_arr = np.array(scores)
    mean_acc = float(scores_arr.mean())
    std_acc = float(scores_arr.std(ddof=1)) if len(scores) > 1 else 0.0

    logger.info(f"k={k} - Mean accuracy over {num_runs} runs: {mean_acc:.4f} +/- {std_acc:.4f} (std)")

    return (mean_acc, std_acc)


def evaluate_fungiclef(args: argparse.Namespace) -> None:
    tic = time.time()

    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")
    logger.info(f"Loading FungiCLEF2023 dataset from {args.dataset_path}")
    dataset = FungiCLEF2023(args.dataset_path)
    metadata_df = _load_fungiclef_metadata(dataset)
    logger.info(f"Loaded metadata for {metadata_df.height} images")

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        x_train, y_train, x_val, y_val = _load_embeddings_with_split(embeddings_path, metadata_df)

        accuracies: dict[int, float] = {}
        accuracies_std: dict[int, float] = {}
        for k in args.k:
            mean_acc, std_acc = _evaluate_single_k(x_train, y_train, x_val, y_val, k, args.runs, args.seed, device)
            accuracies[k] = mean_acc
            accuracies_std[k] = std_acc

        results.append(
            {
                "embeddings_file": str(embeddings_path),
                "method": "knn",
                "num_runs": args.runs,
                "accuracies": accuracies,
                "accuracies_std": accuracies_std,
            }
        )

    _print_summary_table(results, args.k)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("fungiclef.csv")
        _write_results_csv(results, args.k, output_path)

    toc = time.time()
    logger.info(f"FungiCLEF2023 benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "fungiclef",
        allow_abbrev=False,
        help="run FungiCLEF2023 benchmark - 1,604 species classification using KNN",
        description="run FungiCLEF2023 benchmark - 1,604 species classification using KNN",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval fungiclef --embeddings "
            "results/fungiclef_embeddings.parquet --dataset-path ~/Datasets/FungiCLEF2023 --dry-run\n"
            "python -m birder.eval fungiclef --embeddings results/fungiclef/*.parquet "
            "--dataset-path ~/Datasets/FungiCLEF2023 --gpu\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings parquet files"
    )
    subparser.add_argument("--dataset-path", type=str, metavar="PATH", help="path to FungiCLEF2023 dataset root")
    subparser.add_argument(
        "--k", type=int, nargs="+", default=[1, 3], help="k value for k-shot sampling and KNN neighbors"
    )
    subparser.add_argument("--runs", type=int, default=5, help="number of evaluation runs")
    subparser.add_argument("--seed", type=int, default=0, help="base random seed")
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument(
        "--dir", type=str, default="fungiclef", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.dataset_path is None:
        raise cli.ValidationError("--dataset-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_fungiclef(args)



import logging
import unittest

import numpy as np
import torch

from birder.conf import settings
from birder.results.classification import Results
from birder.results.classification import SparseResults
from birder.results.classification import top_k_accuracy_score
from birder.results.detection import Results as DetectionResults

logging.disable(logging.CRITICAL)


class TestClassification(unittest.TestCase):
    def test_top_k_accuracy_score(self) -> None:
        y_true = np.array([0, 0, 2, 1, 1, 3])
        y_pred = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )
        indices = top_k_accuracy_score(y_true, y_pred, top_k=2)
        self.assertEqual(indices, [1, 2, 3])

    def test_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)
        self.assertFalse(results.missing_labels)
        self.assertAlmostEqual(results.accuracy, 1.0 / 6.0)
        self.assertAlmostEqual(results.top_k, 5.0 / 6.0)
        self.assertEqual(results._top_k_indices, [1, 2, 3, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

        # Test confusion matrix
        cnf = results.confusion_matrix
        self.assertSequenceEqual(cnf.shape, (4, 4))
        expected = np.array([[0, 1, 1, 0], [0, 0, 2, 0], [0, 0, 1, 0], [0, 0, 1, 0]], dtype=np.int_)
        self.assertSequenceEqual(cnf.tolist(), expected.tolist())

    def test_results_filter(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)
        filtered_results = results.filter_by_labels([0, 2])

        self.assertEqual(len(filtered_results), 3)
        self.assertFalse(filtered_results.missing_labels)
        self.assertAlmostEqual(filtered_results.accuracy, 1.0 / 3.0)

    def test_most_confused_pairs(self) -> None:
        sample_list = [f"file{i}.jpg" for i in range(9)]
        labels = [0, 0, 0, 0, 1, 1, 1, 2, 2]
        predictions = [1, 1, 1, 0, 1, 1, 0, 2, 1]
        label_names = ["l0", "l1", "l2"]
        output = np.zeros((len(sample_list), len(label_names)), dtype=np.float32)
        for idx, pred in enumerate(predictions):
            output[idx, pred] = 1.0

        results = Results(sample_list, labels, label_names, output)
        pairs = results.most_confused_pairs(n=1)

        self.assertEqual(pairs.height, 1)
        self.assertSequenceEqual(pairs["predicted"].to_list(), ["l1"])
        self.assertSequenceEqual(pairs["actual"].to_list(), ["l0"])
        self.assertSequenceEqual(pairs["amount"].to_list(), [3])
        self.assertSequenceEqual(pairs["reverse"].to_list(), [1])

    def test_partial_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, settings.NO_LABEL, 2, settings.NO_LABEL, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = Results(sample_list, labels, label_names, output)

        self.assertTrue(results.missing_labels)
        self.assertFalse(results.missing_all_labels)
        self.assertEqual(results._valid_length, 4)  # pylint: disable=protected-access
        self.assertAlmostEqual(results.accuracy, 1.0 / 4.0)
        self.assertAlmostEqual(results.top_k, 3.0 / 4.0)
        self.assertEqual(results._top_k_indices, [2, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

    def test_sparse_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)
        self.assertFalse(results.missing_labels)
        self.assertAlmostEqual(results.accuracy, 1.0 / 6.0)
        self.assertAlmostEqual(results.top_k, 5.0 / 6.0)
        self.assertEqual(results._top_k_indices, [1, 2, 3, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])

        cnf = results.confusion_matrix
        self.assertSequenceEqual(cnf.shape, (4, 4))

    def test_sparse_results_filter(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, 0, 2, 1, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)
        filtered_results = results.filter_by_labels([0, 2])

        self.assertEqual(len(filtered_results), 3)
        self.assertFalse(filtered_results.missing_labels)
        self.assertAlmostEqual(filtered_results.accuracy, 1.0 / 3.0)

    def test_partial_sparse_results(self) -> None:
        sample_list = ["file1.jpeg", "file2.jpg", "file3.jpeg", "file4.jpeg", "file5.png", "file6.webp"]
        labels = [0, settings.NO_LABEL, 2, settings.NO_LABEL, 1, 3]
        label_names = ["l0", "l1", "l2", "l3"]
        output = np.array(
            [
                [0.1, 0.25, 0.5, 0.15],
                [0.25, 0.5, 0.15, 0.1],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.25, 0.5, 0.15],
                [0.1, 0.15, 0.5, 0.25],
                [0.1, 0.25, 0.5, 0.15],
            ]
        )

        results = SparseResults(sample_list, labels, label_names, output, sparse_k=3)

        self.assertTrue(results.missing_labels)
        self.assertFalse(results.missing_all_labels)
        self.assertEqual(results._valid_length, 4)  # pylint: disable=protected-access
        self.assertAlmostEqual(results.accuracy, 1.0 / 4.0)
        self.assertAlmostEqual(results.top_k, 3.0 / 4.0)
        self.assertEqual(results._top_k_indices, [2, 4, 5])  # pylint: disable=protected-access
        self.assertEqual(results.predictions.tolist(), [2, 1, 2, 2, 2, 2])
        self.assertEqual(results.prediction_names.to_list(), ["l2", "l1", "l2", "l2", "l2", "l2"])

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [0, 1, 2, 3])


class TestDetectionResults(unittest.TestCase):
    def test_results(self) -> None:
        sample_paths = ["img1.jpg", "img2.jpg"]
        class_to_idx = {"cat": 1, "dog": 2}
        targets = [
            {"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]), "labels": torch.tensor([1])},
            {"boxes": torch.tensor([[5.0, 5.0, 20.0, 20.0]]), "labels": torch.tensor([2])},
        ]
        detections = [
            {
                "boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]),
                "scores": torch.tensor([0.95]),
                "labels": torch.tensor([1]),
            },
            {
                "boxes": torch.tensor([[5.0, 5.0, 20.0, 20.0]]),
                "scores": torch.tensor([0.9]),
                "labels": torch.tensor([2]),
            },
        ]

        results = DetectionResults(sample_paths, targets, detections, class_to_idx)
        self.assertEqual(len(results), 2)
        self.assertEqual(results.label_names, ["Background", "cat", "dog"])
        self.assertAlmostEqual(results.map, 1.0, places=4)

        report = results.detailed_report()
        self.assertSequenceEqual(report["Class"].to_list(), [1, 2])
        self.assertSequenceEqual(report["Objects"].to_list(), [1, 1])

    def test_confusion_matrix(self) -> None:
        sample_paths = ["img1.jpg", "img2.jpg", "img3.jpg"]
        class_to_idx = {"cat": 1, "dog": 2}
        targets = [
            {"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0], [20.0, 20.0, 30.0, 30.0]]), "labels": torch.tensor([1, 2])},
            {"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]), "labels": torch.tensor([1])},
            {"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]), "labels": torch.tensor([2])},
        ]
        detections = [
            {
                "boxes": torch.tensor(
                    [
                        [0.0, 0.0, 10.0, 10.0],
                        [20.0, 20.0, 30.0, 30.0],
                        [40.0, 40.0, 50.0, 50.0],
                    ]
                ),
                "scores": torch.tensor([0.95, 0.9, 0.4]),
                "labels": torch.tensor([1, 2, 1]),
            },
            {
                "boxes": torch.tensor(
                    [
                        [0.0, 0.0, 10.0, 10.0],
                        [0.0, 0.0, 9.0, 9.0],
                        [50.0, 50.0, 60.0, 60.0],
                    ]
                ),
                "scores": torch.tensor([0.9, 0.8, 0.6]),
                "labels": torch.tensor([2, 1, 2]),
            },
            {
                "boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]),
                "scores": torch.tensor([0.4]),
                "labels": torch.tensor([2]),
            },
        ]

        results = DetectionResults(sample_paths, targets, detections, class_to_idx)
        cnf = results.confusion_matrix(score_threshold=0.5, iou_threshold=0.5)
        expected = np.array([[0, 1, 1], [0, 1, 1], [1, 0, 1]], dtype=np.int_)
        self.assertSequenceEqual(cnf.tolist(), expected.tolist())

    def test_confusion_matrix_iou_threshold(self) -> None:
        sample_paths = ["img1.jpg"]
        class_to_idx = {"cat": 1}
        targets = [{"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]), "labels": torch.tensor([1])}]
        detections = [
            {
                "boxes": torch.tensor([[0.0, 0.0, 8.0, 8.0]]),
                "scores": torch.tensor([0.9]),
                "labels": torch.tensor([1]),
            }
        ]

        results = DetectionResults(sample_paths, targets, detections, class_to_idx)
        cnf_low = results.confusion_matrix(score_threshold=0.5, iou_threshold=0.5)
        cnf_high = results.confusion_matrix(score_threshold=0.5, iou_threshold=0.7)

        expected_low = np.array([[0, 0], [0, 1]], dtype=np.int_)
        expected_high = np.array([[0, 1], [1, 0]], dtype=np.int_)
        self.assertSequenceEqual(cnf_low.tolist(), expected_low.tolist())
        self.assertSequenceEqual(cnf_high.tolist(), expected_high.tolist())

    def test_most_confused_pairs(self) -> None:
        sample_paths = ["img1.jpg", "img2.jpg", "img3.jpg", "img4.jpg", "img5.jpg", "img6.jpg"]
        class_to_idx = {"cat": 1, "dog": 2}
        targets = [
            {"boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]), "labels": torch.tensor([1])},
            {"boxes": torch.tensor([[20.0, 20.0, 30.0, 30.0]]), "labels": torch.tensor([1])},
            {"boxes": torch.tensor([[40.0, 40.0, 50.0, 50.0]]), "labels": torch.tensor([1])},
            {"boxes": torch.tensor([[0.0, 20.0, 10.0, 30.0]]), "labels": torch.tensor([2])},
            {"boxes": torch.tensor([[20.0, 0.0, 30.0, 10.0]]), "labels": torch.tensor([2])},
            {
                "boxes": torch.tensor([], dtype=torch.float32).reshape(0, 4),
                "labels": torch.tensor([], dtype=torch.int64),
            },
        ]
        detections = [
            {
                "boxes": torch.tensor([[0.0, 0.0, 10.0, 10.0]]),
                "scores": torch.tensor([0.95]),
                "labels": torch.tensor([2]),
            },
            {
                "boxes": torch.tensor([[20.0, 20.0, 30.0, 30.0]]),
                "scores": torch.tensor([0.9]),
                "labels": torch.tensor([2]),
            },
            {
                "boxes": torch.tensor([[40.0, 40.0, 50.0, 50.0]]),
                "scores": torch.tensor([0.85]),
                "labels": torch.tensor([2]),
            },
            {
                "boxes": torch.tensor([[0.0, 20.0, 10.0, 30.0]]),
                "scores": torch.tensor([0.9]),
                "labels": torch.tensor([1]),
            },
            {
                "boxes": torch.tensor([[20.0, 0.0, 30.0, 10.0]]),
                "scores": torch.tensor([0.85]),
                "labels": torch.tensor([1]),
            },
            {
                "boxes": torch.tensor([[60.0, 60.0, 70.0, 70.0]]),
                "scores": torch.tensor([0.9]),
                "labels": torch.tensor([2]),
            },
        ]

        results = DetectionResults(sample_paths, targets, detections, class_to_idx)
        pairs = results.most_confused_pairs(n=3, score_threshold=0.5, iou_threshold=0.5)
        self.assertEqual(pairs.height, 3)
        self.assertSequenceEqual(pairs["predicted"].to_list(), ["dog", "cat", "dog"])
        self.assertSequenceEqual(pairs["actual"].to_list(), ["cat", "dog", "Background"])
        self.assertSequenceEqual(pairs["amount"].to_list(), [3, 2, 1])
        self.assertSequenceEqual(pairs["reverse"].to_list(), [2, 3, 0])



"""
Transformer Attribution (Gradient-weighted Attention Rollout), adapted from
https://github.com/hila-chefer/Transformer-Explainability

Paper "Transformer Interpretability Beyond Attention Visualization", https://arxiv.org/abs/2012.09838
"""

# Reference license: MIT

from collections.abc import Callable
from pathlib import Path
from typing import Optional

import numpy as np
import torch
from PIL import Image
from torch import nn

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import predict_class
from birder.introspection.base import preprocess_image
from birder.introspection.base import show_mask_on_image
from birder.introspection.base import validate_target_class
from birder.net.vit import Encoder


def compute_attribution_rollout(
    attributions: list[tuple[torch.Tensor, torch.Tensor]], num_special_tokens: int, patch_grid_shape: tuple[int, int]
) -> torch.Tensor:
    """
    NOTE: Uses gradient norm per token instead of element-wise grad * attention multiplication.
    """

    # Assume batch size = 1
    num_tokens = attributions[0][0].size(-1)
    device = attributions[0][0].device

    result = torch.eye(num_tokens, device=device)
    with torch.no_grad():
        for attn_weights, output_grad in attributions:
            # Compute token importance from output gradient norm across embedding dimension
            token_importance = output_grad.norm(dim=-1, keepdim=True)
            token_importance = token_importance.transpose(-1, -2)

            # Weight attention patterns by token importance
            weighted_attn = attn_weights * token_importance.unsqueeze(1)

            # Fuse attention heads and apply non-negativity constraint
            relevance = weighted_attn.mean(dim=1).clamp(min=0)

            # Add residual connection and normalize
            eye = torch.eye(num_tokens, device=device)
            normalized = (relevance + eye) / 2.0
            normalized = normalized / normalized.sum(dim=-1, keepdim=True)

            # Accumulate attention across layers
            result = torch.matmul(normalized, result)

    rollout = result[0]

    if 0 < num_special_tokens:
        source_to_patches = rollout[:num_special_tokens, num_special_tokens:]
        mask = source_to_patches.mean(dim=0)
    else:
        mask = rollout.mean(dim=0)

    mask = mask / (mask.max() + 1e-8)

    grid_h, grid_w = patch_grid_shape
    mask = mask.reshape(grid_h, grid_w)

    return mask


class AttributionGatherer:
    def __init__(self, net: nn.Module, attention_layer_name: str) -> None:
        assert hasattr(net, "encoder") is True and isinstance(net.encoder, Encoder)

        net.encoder.set_need_attn()

        self.net = net
        self.handles: list[torch.utils.hooks.RemovableHandle] = []
        self._gradients: list[torch.Tensor] = []
        self._attention_weights: list[torch.Tensor] = []

        for name, module in self.net.named_modules():
            if name.endswith(attention_layer_name) is True:
                handle = module.register_forward_hook(self._capture_forward)
                self.handles.append(handle)

    def _capture_forward(
        self, _module: nn.Module, _inputs: tuple[torch.Tensor, ...], output: tuple[torch.Tensor, ...] | torch.Tensor
    ) -> None:
        output_tensor = output[0]
        attn_weights = output[1]

        self._attention_weights.append(attn_weights.detach())
        if output_tensor.requires_grad:

            def _store_grad(grad: torch.Tensor) -> None:
                self._gradients.append(grad.detach())

            output_tensor.register_hook(_store_grad)

    def get_captured_data(self) -> list[tuple[torch.Tensor, torch.Tensor]]:
        if len(self._attention_weights) != len(self._gradients):
            raise RuntimeError(
                f"Mismatch between attention weights ({len(self._attention_weights)}) "
                f"and gradients ({len(self._gradients)}). Ensure backward() was called."
            )

        if len(self._attention_weights) == 0:
            raise RuntimeError("No attention data captured. Ensure the model has attention layers.")

        # Pair attention weights with output gradients (gradients reversed to match forward order)
        results = [(attn.cpu(), grad.cpu()) for attn, grad in zip(self._attention_weights, reversed(self._gradients))]

        # Clear storage for next forward pass
        self._gradients = []
        self._attention_weights = []

        return results

    def release(self) -> None:
        for handle in self.handles:
            handle.remove()


class TransformerAttribution:
    def __init__(
        self,
        net: nn.Module,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        attention_layer_name: str = "attn",
    ) -> None:
        self.net = net.eval()
        self.device = device
        self.transform = transform
        self.gatherer = AttributionGatherer(net, attention_layer_name)

    def __call__(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        input_tensor, rgb_img = preprocess_image(image, self.transform, self.device)
        input_tensor.requires_grad_(True)

        self.net.zero_grad()
        logits = self.net(input_tensor)

        if target_class is None:
            target_class = predict_class(logits)
        else:
            validate_target_class(target_class, logits.shape[-1])

        score = logits[0, target_class]
        score.backward()

        attribution_data = self.gatherer.get_captured_data()

        _, _, H, W = input_tensor.shape
        patch_grid_shape = (H // self.net.stem_stride, W // self.net.stem_stride)

        attribution_map = compute_attribution_rollout(
            attribution_data, num_special_tokens=self.net.num_special_tokens, patch_grid_shape=patch_grid_shape
        )

        attribution_img = Image.fromarray(attribution_map.numpy())
        attribution_img = attribution_img.resize((rgb_img.shape[1], rgb_img.shape[0]))
        attribution_arr = np.array(attribution_img)

        visualization = show_mask_on_image(rgb_img, attribution_arr, image_weight=0.4)

        return InterpretabilityResult(
            original_image=rgb_img,
            visualization=visualization,
            raw_output=attribution_arr,
            logits=logits.detach(),
            predicted_class=target_class,
        )

    def __del__(self) -> None:
        if hasattr(self, "gatherer") is True:
            self.gatherer.release()



import logging
from pathlib import Path

import numpy as np
import numpy.typing as npt
import polars as pl

logger = logging.getLogger(__name__)


def l2_normalize(x: npt.NDArray[np.float32], eps: float = 1e-12) -> npt.NDArray[np.float32]:
    norms = np.linalg.norm(x, axis=1, keepdims=True)
    np.maximum(norms, eps, out=norms)

    return x / norms  # type: ignore[no-any-return]


def l2_normalize_(x: npt.NDArray[np.float32], eps: float = 1e-12) -> npt.NDArray[np.float32]:
    norms = np.linalg.norm(x, axis=1, keepdims=True)
    np.maximum(norms, eps, out=norms)
    x /= norms

    return x


def load_embeddings(path: Path | str) -> pl.DataFrame:
    """
    Load embeddings from parquet file as a dataframe

    Auto-detects format:
    - If 'embedding' column exists: use directly
    - If numeric column names (0, 1, 2, ...): treat as logits, convert to array

    Parameters
    ----------
    path
        Path to parquet embedding or logits file.

    Returns
    -------
    DataFrame with columns:
    - id: sample identifier (stem of 'sample' column path)
    - embedding: fixed-size float32 embedding array
    """

    if isinstance(path, str):
        path = Path(path)

    df = pl.read_parquet(path)
    id_expr = pl.col("sample").str.split("/").list.last().str.replace(r"\.[^./\\]+$", "").alias("id")
    if "embedding" in df.columns:
        return df.select(id_expr, pl.col("embedding"))

    # Logits format - numeric column names
    embed_cols = sorted([c for c in df.columns if c.isdigit()], key=int)
    return df.select(
        id_expr,
        pl.concat_arr(pl.col(embed_cols)).cast(pl.Array(pl.Float32, len(embed_cols))).alias("embedding"),
    )



"""
VoVNet v1, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/vovnet.py

Paper "An Energy and GPU-Computation Efficient Backbone Network for Real-Time Object Detection",
https://arxiv.org/abs/1904.09730
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class OSABlock(nn.Module):
    """
    One-Shot Aggregation block
    """

    def __init__(self, in_channels: int, mid_channels: int, out_channels: int, layers_per_block: int) -> None:
        super().__init__()
        self.layers = nn.ModuleList()
        next_in = in_channels
        for _ in range(layers_per_block):
            self.layers.append(
                Conv2dNormActivation(
                    next_in,
                    mid_channels,
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    padding=(1, 1),
                    bias=False,
                )
            )
            next_in = mid_channels

        concat_channels = in_channels + layers_per_block * mid_channels
        self.concat_conv = Conv2dNormActivation(
            concat_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=False,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        features = [x]
        for layer in self.layers:
            x = layer(x)
            features.append(x)

        x = torch.concat(features, dim=1)
        x = self.concat_conv(x)

        return x


class OSAStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        blocks_per_stage: int,
        layers_per_block: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        layers: list[nn.Module] = []
        if downsample is True:
            layers.append(nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(0, 0), ceil_mode=True))

        for idx in range(blocks_per_stage):
            layers.append(
                OSABlock(
                    in_channels if idx == 0 else out_channels,
                    mid_channels,
                    out_channels,
                    layers_per_block,
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.blocks(x)


# pylint: disable=invalid-name
class VoVNet_v1(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        stem_channels: list[int] = self.config["stem_channels"]
        stage_conv_channels: list[int] = self.config["stage_conv_channels"]
        stage_out_channels: list[int] = self.config["stage_out_channels"]
        blocks_per_stage: list[int] = self.config["blocks_per_stage"]
        layers_per_block: int = self.config["layers_per_block"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                stem_channels[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                bias=False,
            ),
            Conv2dNormActivation(
                stem_channels[0],
                stem_channels[1],
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                bias=False,
            ),
            Conv2dNormActivation(
                stem_channels[1],
                stem_channels[2],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                bias=False,
            ),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        in_channels = stem_channels[-1]
        for idx, (mid_ch, out_ch, num_blocks) in enumerate(
            zip(stage_conv_channels, stage_out_channels, blocks_per_stage)
        ):
            stages[f"stage{idx + 1}"] = OSAStage(
                in_channels,
                mid_ch,
                out_ch,
                num_blocks,
                layers_per_block,
                downsample=idx > 0,
            )
            in_channels = out_ch

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = stage_out_channels
        self.embedding_size = stage_out_channels[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad_(False)

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad_(False)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "vovnet_v1_27s",
    VoVNet_v1,
    config={
        "stem_channels": [64, 64, 128],
        "stage_conv_channels": [64, 80, 96, 112],
        "stage_out_channels": [128, 256, 384, 512],
        "blocks_per_stage": [1, 1, 1, 1],
        "layers_per_block": 5,
    },
)
registry.register_model_config(
    "vovnet_v1_39",
    VoVNet_v1,
    config={
        "stem_channels": [64, 64, 128],
        "stage_conv_channels": [128, 160, 192, 224],
        "stage_out_channels": [256, 512, 768, 1024],
        "blocks_per_stage": [1, 1, 2, 2],
        "layers_per_block": 5,
    },
)
registry.register_model_config(
    "vovnet_v1_57",
    VoVNet_v1,
    config={
        "stem_channels": [64, 64, 128],
        "stage_conv_channels": [128, 160, 192, 224],
        "stage_out_channels": [256, 512, 768, 1024],
        "blocks_per_stage": [1, 1, 4, 3],
        "layers_per_block": 5,
    },
)

registry.register_weights(
    "vovnet_v1_27s_il-common",
    {
        "description": "VoVNet v1 27s model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 12.3,
                "sha256": "c119b61fc3ba69e0184cb5b826b66897d55678b9250e206446badee377b1e34f",
            }
        },
        "net": {"network": "vovnet_v1_27s", "tag": "il-common"},
    },
)



"""
Fit YOLO-style anchor boxes using k-means based on COCO-format annotations.

Generated by gpt-5.2-codex xhigh.
"""

import argparse
import json
import logging
import math
from pathlib import Path
from pprint import pformat
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypedDict

import torch

from birder.common import cli
from birder.conf import settings

logger = logging.getLogger(__name__)


class AnchorPreset(TypedDict):
    num_scales: int
    num_anchors: int
    default_size: tuple[int, int]
    format: Literal["pixels", "grid"]


MODEL_PRESETS: dict[str, AnchorPreset] = {
    "yolo_v2": {"num_scales": 1, "num_anchors": 5, "default_size": (416, 416), "format": "grid"},
    "yolo_v3": {"num_scales": 3, "num_anchors": 9, "default_size": (416, 416), "format": "pixels"},
    "yolo_v4": {"num_scales": 3, "num_anchors": 9, "default_size": (608, 608), "format": "pixels"},
    "yolo_v4_tiny": {"num_scales": 2, "num_anchors": 6, "default_size": (416, 416), "format": "pixels"},
}


def _load_ignore_list(ignore_file: Optional[str]) -> set[str]:
    if ignore_file is None:
        return set()

    with open(ignore_file, "r", encoding="utf-8") as handle:
        return {line.strip() for line in handle if line.strip()}


def _load_coco_boxes(
    coco_json_path: str, target_size: tuple[int, int], ignore_list: set[str], min_size: float, ignore_crowd: bool
) -> tuple[torch.Tensor, dict[str, int]]:
    coco_path = Path(coco_json_path)
    if coco_path.exists() is False:
        raise ValueError(f"COCO json not found at {coco_path}")

    with open(coco_path, "r", encoding="utf-8") as handle:
        data = json.load(handle)

    images = {}
    for image in data.get("images", []):
        image_id = image.get("id")
        if image_id is None:
            continue
        images[image_id] = (image.get("width"), image.get("height"), image.get("file_name", ""))

    stats = {
        "total_annotations": 0,
        "used_annotations": 0,
        "crowd_annotations": 0,
        "invalid_bbox": 0,
        "ignored_images": 0,
        "missing_images": 0,
        "missing_size": 0,
        "too_small": 0,
    }
    boxes: list[tuple[float, float]] = []
    target_h = float(target_size[0])
    target_w = float(target_size[1])
    for annotation in data.get("annotations", []):
        stats["total_annotations"] += 1
        if ignore_crowd is True and annotation.get("iscrowd", 0) == 1:
            stats["crowd_annotations"] += 1
            continue

        bbox = annotation.get("bbox")
        if bbox is None or len(bbox) != 4:
            stats["invalid_bbox"] += 1
            continue

        image_id = annotation.get("image_id")
        if image_id not in images:
            stats["missing_images"] += 1
            continue

        img_w, img_h, file_name = images[image_id]
        if file_name in ignore_list:
            stats["ignored_images"] += 1
            continue

        if img_w in {None, 0} or img_h in {None, 0}:
            stats["missing_size"] += 1
            continue

        bbox_w = float(bbox[2])
        bbox_h = float(bbox[3])
        if bbox_w <= 0.0 or bbox_h <= 0.0:
            stats["invalid_bbox"] += 1
            continue

        scaled_w = bbox_w / float(img_w) * target_w
        scaled_h = bbox_h / float(img_h) * target_h
        if scaled_w < min_size or scaled_h < min_size:
            stats["too_small"] += 1
            continue

        boxes.append((scaled_w, scaled_h))

    stats["used_annotations"] = len(boxes)
    if len(boxes) == 0:
        raise ValueError("No valid bounding boxes found for anchor fitting")

    return (torch.tensor(boxes, dtype=torch.float32), stats)


def _wh_iou(boxes: torch.Tensor, anchors: torch.Tensor) -> torch.Tensor:
    boxes = boxes[:, None, :]
    anchors = anchors[None, :, :]
    inter = torch.min(boxes, anchors).prod(dim=2)
    union = boxes.prod(dim=2) + anchors.prod(dim=2) - inter
    return inter / (union + 1e-9)


def _kmeans_plusplus_init(boxes: torch.Tensor, num_anchors: int, generator: torch.Generator) -> torch.Tensor:
    n_local_trials = 2 + int(math.log(num_anchors))

    anchors = []

    first_idx = torch.randint(0, boxes.size(0), (1,), generator=generator).item()
    anchors.append(boxes[first_idx])
    for _ in range(num_anchors - 1):
        anchors_tensor = torch.stack(anchors)
        ious = _wh_iou(boxes, anchors_tensor)
        max_ious = ious.max(dim=1).values

        min_distances = 1.0 - max_ious
        squared_distances = min_distances**2

        probs = squared_distances / (squared_distances.sum() + 1e-9)
        cumulative_probs = torch.cumsum(probs, dim=0)
        r = torch.rand(n_local_trials, generator=generator)
        candidate_indices = torch.searchsorted(cumulative_probs, r, right=True)

        candidate_indices = torch.clamp(candidate_indices, 0, boxes.size(0) - 1)
        candidate_boxes = boxes[candidate_indices]
        candidate_ious = _wh_iou(boxes, candidate_boxes)  # (n_boxes, n_trials)
        candidate_distances = 1.0 - candidate_ious

        min_distances_expanded = min_distances.unsqueeze(1)
        candidate_potentials = torch.min(min_distances_expanded, candidate_distances).sum(dim=0)

        best_trial_idx = torch.argmin(candidate_potentials).item()
        best_idx = candidate_indices[best_trial_idx].item()
        anchors.append(boxes[best_idx])

    return torch.stack(anchors)


def _kmeans_anchors(
    boxes: torch.Tensor, num_anchors: int, seed: Optional[int], max_iter: int
) -> tuple[torch.Tensor, torch.Tensor]:
    if boxes.size(0) < num_anchors:
        raise ValueError(
            f"Not enough boxes ({boxes.size(0)}) to fit {num_anchors} anchors, Reduce --num-anchors or add more data"
        )

    generator = torch.Generator()
    if seed is not None:
        generator.manual_seed(seed)

    anchors = _kmeans_plusplus_init(boxes, num_anchors, generator)
    assignments = torch.full((boxes.size(0),), -1, dtype=torch.int64)

    for _ in range(max_iter):
        ious = _wh_iou(boxes, anchors)
        new_assignments = torch.argmax(ious, dim=1)
        if torch.equal(assignments, new_assignments):
            break

        assignments = new_assignments
        for idx in range(num_anchors):
            mask = assignments == idx
            if mask.any():
                anchors[idx] = boxes[mask].median(dim=0).values
            else:
                rand_idx = torch.randint(0, boxes.size(0), (1,), generator=generator).item()
                anchors[idx] = boxes[rand_idx]

    return (anchors, assignments)


def _format_anchor_groups(anchor_groups: list[torch.Tensor], precision: int) -> list[list[tuple[float, float]]]:
    formatted: list[list[tuple[float, float]]] = []
    for group in anchor_groups:
        formatted.append([(round(float(anchor[0]), precision), round(float(anchor[1]), precision)) for anchor in group])

    return formatted


def _validate_args(
    args: argparse.Namespace,
) -> tuple[tuple[int, int], int, int, Literal["pixels", "grid"], list[float]]:
    preset = MODEL_PRESETS.get(args.preset) if args.preset is not None else None
    size = cli.parse_size(args.size) if args.size is not None else (preset["default_size"] if preset else None)
    if size is None:
        raise cli.ValidationError("Missing --size. Provide --size or use a --preset")

    num_scales = args.num_scales if args.num_scales is not None else (preset["num_scales"] if preset else None)
    num_anchors = args.num_anchors if args.num_anchors is not None else (preset["num_anchors"] if preset else None)
    output_format = args.format if args.format is not None else (preset["format"] if preset else None)
    if num_scales is None or num_anchors is None or output_format is None:
        raise cli.ValidationError(
            "Missing configuration. Provide --num-scales, --num-anchors and --format or use a --preset"
        )
    if num_scales < 1:
        raise cli.ValidationError("--num-scales must be >= 1")
    if num_anchors < 1:
        raise cli.ValidationError("--num-anchors must be >= 1")
    if num_anchors % num_scales != 0:
        raise cli.ValidationError("--num-anchors must be divisible by --num-scales")

    strides: list[float] = []
    if output_format == "grid":
        if args.stride is None:
            raise cli.ValidationError("--format grid requires --stride values per scale")

        strides = [float(value) for value in args.stride]
        if len(strides) != num_scales:
            raise cli.ValidationError("--stride must provide one value per scale when --format grid is used")
        if any(value <= 0 for value in strides):
            raise cli.ValidationError("--stride values must be > 0")

    return (size, num_scales, num_anchors, output_format, strides)


# pylint: disable=too-many-locals
def auto_anchors(args: argparse.Namespace) -> None:
    size, num_scales, num_anchors, output_format, strides = _validate_args(args)

    ignore_list = _load_ignore_list(args.ignore_file)
    boxes, stats = _load_coco_boxes(
        args.coco_json_path, size, ignore_list, args.min_size, ignore_crowd=not args.include_crowd
    )

    if args.preset is not None:
        logger.info(f"Using preset {args.preset}")

    logger.info(f"Fitting anchors using size={size[0]}x{size[1]}")
    logger.info(
        f"Annotations: total={stats['total_annotations']}, used={stats['used_annotations']}, "
        f"crowd={stats['crowd_annotations']}, invalid={stats['invalid_bbox']}, "
        f"ignored={stats['ignored_images']}, missing={stats['missing_images']}, "
        f"missing_size={stats['missing_size']}, too_small={stats['too_small']}"
    )

    anchors, _assignments = _kmeans_anchors(boxes, num_anchors, args.seed, args.max_iter)
    areas = anchors.prod(dim=1)
    anchors = anchors[torch.argsort(areas)]
    anchors_per_scale = num_anchors // num_scales
    anchor_groups = [anchors[i : i + anchors_per_scale] for i in range(0, num_anchors, anchors_per_scale)]

    ious = _wh_iou(boxes, anchors)
    best_iou = ious.max(dim=1).values
    logger.info(f"Mean IoU: {best_iou.mean().item():.4f}")

    formatted_groups = _format_anchor_groups(anchor_groups, args.precision)
    anchors_output = None
    if output_format == "pixels":
        if num_scales == 1:
            formatted_anchors: Any = formatted_groups[0]
        else:
            formatted_anchors = formatted_groups

        print("Anchors (pixels):")
        print(pformat(formatted_anchors))
        anchors_output = formatted_anchors

    if output_format == "grid":
        grid_groups: list[torch.Tensor] = []
        for group, stride in zip(anchor_groups, strides):
            grid_group = group.clone()
            grid_group[:, 0] = grid_group[:, 0] / stride
            grid_group[:, 1] = grid_group[:, 1] / stride
            grid_groups.append(grid_group)

        formatted_grid = _format_anchor_groups(grid_groups, args.precision)
        if num_scales == 1:
            formatted_grid_output: Any = formatted_grid[0]
        else:
            formatted_grid_output = formatted_grid

        print("Anchors (grid units):")
        print(pformat(formatted_grid_output))
        anchors_output = formatted_grid_output

    if args.output is not None:
        payload = {
            "anchors": anchors_output,
            "format": output_format,
            "size": [size[0], size[1]],
        }
        if output_format == "grid":
            payload["strides"] = strides

        with open(args.output, "w", encoding="utf-8") as handle:
            json.dump(payload, handle, indent=2)

        logger.info(f"Wrote anchors to {args.output}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "auto-anchors",
        allow_abbrev=False,
        help="fit YOLO anchors with k-means on a COCO dataset",
        description="fit YOLO anchors with k-means on a COCO dataset",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools auto-anchors --preset yolo_v4 --size 640 "
            "--coco-json-path data/detection_data/training_annotations_coco.json\n"
            "python -m birder.tools auto-anchors --size 640 --num-anchors 9 --num-scales 3 --format pixels "
            "--coco-json-path data/detection_data/training_annotations_coco.json\n"
            "python -m birder.tools auto-anchors --preset yolo_v4_tiny --size 416 416 "
            "--coco-json-path ~/Datasets/cocodataset/annotations/instances_train2017.json --output anchors.json\n"
            "python -m birder.tools auto-anchors --preset yolo_v2 --stride 32 "
            "--coco-json-path data/detection_data/training_annotations_coco.json\n"
            "python -m birder.tools auto-anchors --size 640 --num-anchors 9 --num-scales 3 "
            "--format grid --stride 8 16 32 --coco-json-path data/detection_data/training_annotations_coco.json\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--preset", type=str, choices=sorted(MODEL_PRESETS.keys()), help="YOLO preset for anchor formatting"
    )
    subparser.add_argument(
        "--size",
        type=int,
        nargs="+",
        metavar=("H", "W"),
        help="target image size as [height, width], required without --preset",
    )
    subparser.add_argument("--num-anchors", type=int, help="number of anchors to fit, required without --preset")
    subparser.add_argument("--num-scales", type=int, help="number of output scales, required without --preset")
    subparser.add_argument(
        "--format", type=str, choices=["pixels", "grid"], help="anchor output format, required without --preset"
    )
    subparser.add_argument(
        "--stride", type=int, nargs="+", default=[32], help="strides per scale used to convert anchors to grid units"
    )
    subparser.add_argument(
        "--min-size", type=float, default=1.0, help="minimum scaled box size to include in anchor fitting"
    )
    subparser.add_argument("--include-crowd", default=False, action="store_true", help="include crowd annotations")
    subparser.add_argument("--seed", type=int, help="random seed for k-means initialization")
    subparser.add_argument("--max-iter", type=int, default=1000, help="maximum k-means iterations")
    subparser.add_argument("--precision", type=int, default=1, help="number of decimals to keep in anchor output")
    subparser.add_argument(
        "--ignore-file", type=str, metavar="FILE", help="file containing image names to skip (one per line)"
    )
    subparser.add_argument(
        "--coco-json-path",
        type=str,
        default=f"{settings.TRAINING_DETECTION_ANNOTATIONS_PATH}_coco.json",
        help="training COCO json path",
    )
    subparser.add_argument("--output", type=str, help="write anchors as JSON to this path")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    auto_anchors(args)



"""
RoPE ViT, adapted from
https://github.com/naver-ai/rope-vit/blob/main/deit/models_v2_rope.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/pos_embed_sincos.py

Paper "Rotary Position Embedding for Vision Transformer", https://arxiv.org/abs/2403.13298

Changes from original:
* Implemented only axial RoPE (EVA style RoPE)
"""

# Reference license: Apache-2.0 (both)

import math
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import FFN
from birder.layers import LayerScale
from birder.layers import MultiHeadAttentionPool
from birder.layers import SwiGLU_FFN
from birder.layers.activations import get_activation_module
from birder.model_registry import registry
from birder.net._rope_vit_configs import register_rope_vit_configs
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.base import TokenRetentionResultType
from birder.net.base import normalize_out_indices
from birder.net.vit import PatchEmbed
from birder.net.vit import adjust_position_embedding


def build_rotary_pos_embed(
    dim: int,
    temperature: float,
    grid_size: tuple[int, int],
    grid_indexing: str,
    grid_offset: int,
    pt_grid_size: Optional[tuple[int, int]],
    device: Optional[torch.device] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    assert dim % 4 == 0
    num_bands = dim // 4
    exp = torch.arange(0, num_bands, 1, device=device) / num_bands
    bands = 1.0 / (temperature**exp)

    if pt_grid_size is None:
        pt_grid_size = grid_size

    t = [(torch.arange(s, device=device) + grid_offset) / s * p for s, p in zip(grid_size, pt_grid_size)]
    grid = torch.stack(torch.meshgrid(t, indexing=grid_indexing), dim=-1)
    grid = grid.unsqueeze(-1)
    pos = grid * bands
    sin_emb = pos.sin()
    cos_emb = pos.cos()

    num_spatial_dim = grid_size[0] * grid_size[1]

    sin_emb = sin_emb.reshape(num_spatial_dim, -1).repeat_interleave(2, -1)
    cos_emb = cos_emb.reshape(num_spatial_dim, -1).repeat_interleave(2, -1)

    return (sin_emb, cos_emb)


def rotate_half(x: torch.Tensor) -> torch.Tensor:
    # Taken from: https://github.com/facebookresearch/capi/blob/main/model.py
    x1, x2 = x.chunk(2, dim=-1)
    return torch.concat((-x2, x1), dim=-1)


def rotate_half_interleaved(x: torch.Tensor) -> torch.Tensor:
    return torch.stack([-x[..., 1::2], x[..., ::2]], dim=-1).reshape(x.size())


def apply_rotary_pos_embed(x: torch.Tensor, embed: torch.Tensor) -> torch.Tensor:
    sin_emb, cos_emb = embed.tensor_split(2, dim=-1)
    if cos_emb.ndim == 3:
        return x * cos_emb.unsqueeze(1).expand_as(x) + rotate_half(x) * sin_emb.unsqueeze(1).expand_as(x)

    return x * cos_emb + rotate_half(x) * sin_emb


def apply_interleaved_rotary_pos_embed(x: torch.Tensor, embed: torch.Tensor) -> torch.Tensor:
    sin_emb, cos_emb = embed.tensor_split(2, dim=-1)
    if cos_emb.ndim == 3:
        return x * cos_emb.unsqueeze(1).expand_as(x) + rotate_half_interleaved(x) * sin_emb.unsqueeze(1).expand_as(x)

    return x * cos_emb + rotate_half_interleaved(x) * sin_emb


class SequentialWithRope(nn.Sequential):
    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:  # pylint: disable=arguments-differ
        for module in self:
            x = module(x, rope)

        return x


class RoPE(nn.Module):
    def __init__(
        self,
        dim: int,
        temperature: float,
        grid_size: tuple[int, int],
        grid_indexing: Literal["ij", "xy"],
        grid_offset: int,
        pt_grid_size: Optional[tuple[int, int]] = None,
        rope_rot_type: str = "standard",
        device: Optional[torch.device] = None,
    ) -> None:
        super().__init__()
        if rope_rot_type == "standard":
            self.apply_fn = apply_rotary_pos_embed
        elif rope_rot_type == "interleaved":
            self.apply_fn = apply_interleaved_rotary_pos_embed
        else:
            raise ValueError(f"Unknown rope_rot_type, got '{rope_rot_type}'")

        sin_emb, cos_emb = build_rotary_pos_embed(
            dim,
            temperature,
            grid_size=grid_size,
            grid_indexing=grid_indexing,
            grid_offset=grid_offset,
            pt_grid_size=pt_grid_size,
            device=device,
        )
        self.pos_embed = nn.Buffer(torch.concat((sin_emb, cos_emb), dim=-1), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.apply_fn(x, self.pos_embed)


class RoPEAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        attn_drop: float,
        proj_drop: float,
        num_special_tokens: int,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"

        self.is_causal = False
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5
        self.num_special_tokens = num_special_tokens
        if rope_rot_type == "standard":
            self.apply_rot_fn = apply_rotary_pos_embed
        elif rope_rot_type == "interleaved":
            self.apply_rot_fn = apply_interleaved_rotary_pos_embed
        else:
            raise ValueError(f"Unknown rope_rot_type, got '{rope_rot_type}'")

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        if qk_norm is True:
            self.q_norm = norm_layer(self.head_dim, eps=norm_layer_eps)
            self.k_norm = norm_layer(self.head_dim, eps=norm_layer_eps)
        else:
            self.q_norm = nn.Identity()
            self.k_norm = nn.Identity()

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        B, N, C = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q = self.q_norm(q)
        k = self.k_norm(k)

        n = self.num_special_tokens
        q = torch.concat([q[:, :, :n, :], self.apply_rot_fn(q[:, :, n:, :], rope)], dim=2)
        k = torch.concat([k[:, :, :n, :], self.apply_rot_fn(k[:, :, n:, :], rope)], dim=2)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, is_causal=self.is_causal, scale=self.scale
        )

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class EncoderBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: Optional[int],
        num_special_tokens: int,
        dropout: float,
        attention_dropout: float,
        drop_path: float,
        activation_layer: Callable[..., nn.Module],
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()

        if mlp_dim is None:
            mlp_dim = hidden_dim * 4

        # Attention block
        self.norm1 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.attn = RoPEAttention(
            hidden_dim,
            num_heads,
            attn_drop=attention_dropout,
            proj_drop=dropout,
            num_special_tokens=num_special_tokens,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            rope_rot_type=rope_rot_type,
        )
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.norm2 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.mlp = mlp_layer(hidden_dim, mlp_dim, act_layer=activation_layer, dropout=dropout)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x), rope)))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x))))

        return x

    def set_causal_attention(self, is_causal: bool = True) -> None:
        self.attn.is_causal = is_causal


class Encoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        num_heads: int,
        hidden_dim: int,
        mlp_dim: int,
        num_special_tokens: int,
        dropout: float,
        attention_dropout: float,
        dpr: list[float],
        pre_norm: bool = False,
        qkv_bias: bool = True,
        qk_norm: bool = False,
        activation_layer: Callable[..., nn.Module] = nn.GELU,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        pre_layers = []
        if dropout > 0.0:
            pre_layers.append(nn.Dropout(dropout))
        if pre_norm is True:
            pre_layers.append(norm_layer(hidden_dim, eps=norm_layer_eps))

        self.pre_block = nn.Sequential(*pre_layers)

        layers = []
        for i in range(num_layers):
            layers.append(
                EncoderBlock(
                    num_heads,
                    hidden_dim,
                    mlp_dim,
                    num_special_tokens,
                    dropout,
                    attention_dropout,
                    dpr[i],
                    activation_layer=activation_layer,
                    layer_scale_init_value=layer_scale_init_value,
                    norm_layer=norm_layer,
                    norm_layer_eps=norm_layer_eps,
                    mlp_layer=mlp_layer,
                    qkv_bias=qkv_bias,
                    qk_norm=qk_norm,
                    rope_rot_type=rope_rot_type,
                )
            )

        self.block = SequentialWithRope(*layers)

    def forward(self, x: torch.Tensor, rope: torch.Tensor) -> torch.Tensor:
        x = self.pre_block(x)
        return self.block(x, rope)

    def forward_features(
        self, x: torch.Tensor, rope: torch.Tensor, out_indices: Optional[list[int]] = None
    ) -> list[torch.Tensor]:
        x = self.pre_block(x)

        out_indices_set = set(out_indices) if out_indices is not None else None
        xs = []
        for idx, blk in enumerate(self.block):
            x = blk(x, rope)
            if out_indices_set is None or idx in out_indices_set:
                xs.append(x)

        return xs

    def set_causal_attention(self, is_causal: bool = True) -> None:
        for b in self.block:
            b.set_causal_attention(is_causal)


class MAEDecoderBlock(nn.Module):
    def __init__(
        self,
        num_heads: int,
        hidden_dim: int,
        num_special_tokens: int,
        activation_layer: Callable[..., nn.Module],
        grid_size: tuple[int, int],
        rope_grid_indexing: Literal["ij", "xy"],
        rope_grid_offset: int,
        rope_temperature: float,
        layer_scale_init_value: Optional[float] = None,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        norm_layer_eps: float = 1e-6,
        mlp_layer: Callable[..., nn.Module] = FFN,
        rope_rot_type: str = "standard",
    ) -> None:
        super().__init__()
        mlp_dim = hidden_dim * 4
        self.rope = RoPE(
            hidden_dim // num_heads,
            temperature=rope_temperature,
            grid_size=grid_size,
            grid_indexing=rope_grid_indexing,
            grid_offset=rope_grid_offset,
            rope_rot_type=rope_rot_type,
        )

        # Attention block
        self.norm1 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.attn = RoPEAttention(
            hidden_dim,
            num_heads,
            attn_drop=0.0,
            proj_drop=0.0,
            num_special_tokens=num_special_tokens,
            rope_rot_type=rope_rot_type,
        )
        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()

        # MLP block
        self.norm2 = norm_layer(hidden_dim, eps=norm_layer_eps)
        self.mlp = mlp_layer(hidden_dim, mlp_dim, act_layer=activation_layer, dropout=0.0)
        if layer_scale_init_value is not None:
            self.layer_scale_2 = LayerScale(hidden_dim, layer_scale_init_value)
        else:
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.layer_scale_1(self.attn(self.norm1(x), self.rope.pos_embed))
        x = x + self.layer_scale_2(self.mlp(self.norm2(x)))

        return x


# pylint: disable=invalid-name,too-many-instance-attributes
class RoPE_ViT(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin, MaskedTokenRetentionMixin):
    block_group_regex = r"encoder\.block\.(\d+)"

    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        attention_dropout = 0.0
        dropout = 0.0
        pos_embed_special_tokens: bool = self.config.get("pos_embed_special_tokens", True)
        patch_size: int = self.config["patch_size"]
        num_layers: int = self.config["num_layers"]
        num_heads: int = self.config["num_heads"]
        hidden_dim: int = self.config["hidden_dim"]
        mlp_dim: int = self.config["mlp_dim"]
        layer_scale_init_value: Optional[float] = self.config.get("layer_scale_init_value", None)
        pre_norm: bool = self.config.get("pre_norm", False)
        post_norm: bool = self.config.get("post_norm", True)
        qkv_bias: bool = self.config.get("qkv_bias", True)
        qk_norm: bool = self.config.get("qk_norm", False)
        num_reg_tokens: int = self.config.get("num_reg_tokens", 0)
        class_token: bool = self.config.get("class_token", True)
        attn_pool_head: bool = self.config.get("attn_pool_head", False)
        attn_pool_num_heads: Optional[int] = self.config.get("attn_pool_num_heads", None)
        attn_pool_special_tokens: bool = self.config.get("attn_pool_special_tokens", False)
        norm_layer_type: str = self.config.get("norm_layer_type", "LayerNorm")
        norm_layer_eps: float = self.config.get("norm_layer_eps", 1e-6)
        mlp_layer_type: str = self.config.get("mlp_layer_type", "FFN")
        act_layer_type: Optional[str] = self.config.get("act_layer_type", None)  # Default according to mlp type
        out_indices: Optional[list[int]] = self.config.get("out_indices", None)
        rope_rot_type: Literal["standard", "interleaved"] = self.config.get("rope_rot_type", "standard")
        rope_grid_indexing: Literal["ij", "xy"] = self.config.get("rope_grid_indexing", "ij")
        rope_grid_offset: int = self.config.get("rope_grid_offset", 0)
        rope_temperature: float = self.config.get("rope_temperature", 100.0)
        pt_grid_size: Optional[tuple[int, int]] = self.config.get("pt_grid_size", None)
        drop_path_rate: float = self.config["drop_path_rate"]

        if norm_layer_type == "LayerNorm":
            norm_layer = nn.LayerNorm
        elif norm_layer_type == "RMSNorm":
            norm_layer = nn.RMSNorm
        else:
            raise ValueError(f"Unknown norm_layer_type '{norm_layer_type}'")

        if mlp_layer_type == "FFN":
            mlp_layer = FFN
            act_layer = nn.GELU
        elif mlp_layer_type == "SwiGLU_FFN":
            mlp_layer = SwiGLU_FFN
            act_layer = nn.SiLU
        else:
            raise ValueError(f"Unknown mlp_layer_type '{mlp_layer_type}'")

        if act_layer_type is not None:
            act_layer = get_activation_module(act_layer_type)

        torch._assert(image_size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(image_size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(hidden_dim % num_heads == 0, "Hidden dim indivisible by num heads!")
        self.pos_embed_special_tokens = pos_embed_special_tokens
        self.patch_size = patch_size
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.layer_scale_init_value = layer_scale_init_value
        self.num_reg_tokens = num_reg_tokens
        self.attn_pool_special_tokens = attn_pool_special_tokens
        self.norm_layer = norm_layer
        self.norm_layer_eps = norm_layer_eps
        self.mlp_layer = mlp_layer
        self.act_layer = act_layer
        self.out_indices = normalize_out_indices(out_indices, num_layers)
        self.rope_rot_type = rope_rot_type
        self.rope_grid_indexing = rope_grid_indexing
        self.rope_grid_offset = rope_grid_offset
        self.rope_temperature = rope_temperature

        # Cast in case config was loaded from a json (no tuples),
        # TorchScript does not accept a list when tuple expected
        if isinstance(pt_grid_size, list):
            pt_grid_size = tuple(pt_grid_size)  # type: ignore[unreachable]

        self.pt_grid_size = pt_grid_size
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, num_layers)]  # Stochastic depth decay rule

        self.conv_proj = nn.Conv2d(
            self.input_channels,
            hidden_dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
            bias=not pre_norm,
        )
        self.patch_embed = PatchEmbed()

        seq_length = (image_size[0] // patch_size) * (image_size[1] // patch_size)
        self.num_special_tokens = 0

        # Add a class token
        if class_token is True:
            self.class_token = nn.Parameter(torch.zeros(1, 1, hidden_dim))
            self.num_special_tokens += 1
            if pos_embed_special_tokens is True:
                seq_length += 1
        else:
            self.class_token = None

        # Add optional register tokens
        if self.num_reg_tokens > 0:
            self.reg_tokens = nn.Parameter(torch.zeros(1, self.num_reg_tokens, hidden_dim))
            self.num_special_tokens += self.num_reg_tokens
            if pos_embed_special_tokens is True:
                seq_length += self.num_reg_tokens
        else:
            self.reg_tokens = None

        # Add positional embedding
        self.pos_embedding = nn.Parameter(torch.empty(1, seq_length, hidden_dim).normal_(std=0.02))

        # RoPE
        self.rope = RoPE(
            hidden_dim // num_heads,
            temperature=self.rope_temperature,
            grid_size=(image_size[0] // patch_size, image_size[1] // patch_size),
            grid_indexing=rope_grid_indexing,
            grid_offset=rope_grid_offset,
            pt_grid_size=self.pt_grid_size,
            rope_rot_type=rope_rot_type,
        )

        # Encoder
        self.encoder = Encoder(
            num_layers,
            num_heads,
            hidden_dim,
            mlp_dim,
            self.num_special_tokens,
            dropout,
            attention_dropout,
            dpr,
            pre_norm=pre_norm,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            activation_layer=act_layer,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            mlp_layer=mlp_layer,
            rope_rot_type=rope_rot_type,
        )

        if post_norm is True:
            self.norm = norm_layer(hidden_dim, eps=norm_layer_eps)
        else:
            self.norm = nn.Identity()

        if attn_pool_head is False:
            self.attn_pool = None
        else:
            if attn_pool_num_heads is None:
                attn_pool_num_heads = num_heads

            self.attn_pool = MultiHeadAttentionPool(hidden_dim, attn_pool_num_heads, mlp_dim, qkv_bias=True)

        num_return_stages = len(self.out_indices) if self.out_indices is not None else 1
        self.return_stages = [f"stage{stage_idx + 1}" for stage_idx in range(num_return_stages)]
        self.return_channels = [hidden_dim] * num_return_stages
        self.embedding_size = hidden_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = hidden_dim
        self.encoding_size = hidden_dim
        self.decoder_block = partial(
            MAEDecoderBlock,
            16,
            num_special_tokens=self.num_special_tokens,
            activation_layer=act_layer,
            grid_size=(image_size[0] // patch_size, image_size[1] // patch_size),
            rope_grid_indexing=rope_grid_indexing,
            rope_grid_offset=rope_grid_offset,
            rope_temperature=rope_temperature,
            layer_scale_init_value=layer_scale_init_value,
            norm_layer=norm_layer,
            norm_layer_eps=norm_layer_eps,
            mlp_layer=mlp_layer,
            rope_rot_type=rope_rot_type,
        )

        # Weight initialization
        if isinstance(self.conv_proj, nn.Conv2d):
            # Init the patchify stem
            fan_in = self.conv_proj.in_channels * self.conv_proj.kernel_size[0] * self.conv_proj.kernel_size[1]
            nn.init.trunc_normal_(self.conv_proj.weight, std=math.sqrt(1 / fan_in))
            if self.conv_proj.bias is not None:
                nn.init.zeros_(self.conv_proj.bias)

        if isinstance(self.classifier, nn.Linear):
            nn.init.zeros_(self.classifier.weight)
            nn.init.zeros_(self.classifier.bias)

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.pos_embedding

        if H == self.size[0] and W == self.size[1]:
            return self.pos_embedding

        return adjust_position_embedding(
            self.pos_embedding,
            (self.size[0] // self.patch_size, self.size[1] // self.patch_size),
            (H // self.patch_size, W // self.patch_size),
            self.num_special_tokens if self.pos_embed_special_tokens is True else 0,
            antialias=False,
        )

    def _get_rope_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False:
            return self.rope.pos_embed

        if H == self.size[0] and W == self.size[1]:
            return self.rope.pos_embed

        return torch.concat(
            build_rotary_pos_embed(
                self.hidden_dim // self.num_heads,
                self.rope_temperature,
                grid_size=(H // self.patch_size, W // self.patch_size),
                grid_indexing=self.rope_grid_indexing,
                grid_offset=self.rope_grid_offset,
                pt_grid_size=self.pt_grid_size,
            ),
            dim=-1,
        ).to(self.rope.pos_embed.device, dtype=self.rope.pos_embed.dtype)

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad_(False)

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad_(True)

        if unfreeze_features is True:
            if self.attn_pool is not None:
                for param in self.attn_pool.parameters():
                    param.requires_grad_(True)

    def set_causal_attention(self, is_causal: bool = True) -> None:
        self.encoder.set_causal_attention(is_causal)

    def transform_to_backbone(self) -> None:
        super().transform_to_backbone()
        self.norm = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        H, W = x.shape[-2:]
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        rope = self._get_rope_embed(H, W)
        if self.out_indices is None:
            xs = [self.encoder(x, rope)]
        else:
            xs = self.encoder.forward_features(x, rope, out_indices=self.out_indices)

        out: dict[str, torch.Tensor] = {}
        for stage_name, stage_x in zip(self.return_stages, xs, strict=True):
            stage_x = stage_x[:, self.num_special_tokens :]
            stage_x = stage_x.permute(0, 2, 1)
            B, C, _ = stage_x.size()
            stage_x = stage_x.reshape(B, C, H // self.patch_size, W // self.patch_size)
            out[stage_name] = stage_x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.conv_proj.parameters():
            param.requires_grad_(False)

        self.pos_embedding.requires_grad_(False)

        for idx, module in enumerate(self.encoder.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad_(False)

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        H, W = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        # Add pos embedding without special tokens
        pos_embedding = self._get_pos_embed(H, W)
        if self.pos_embed_special_tokens is True:
            x = x + pos_embedding[:, self.num_special_tokens :, :]
        else:
            x = x + pos_embedding

        # Mask tokens
        if ids_keep is not None:
            x = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, x.size(2)))

            rope_dim = self.rope.pos_embed.size(1)
            rope = self.rope.pos_embed.unsqueeze(0).repeat(x.size(0), 1, 1)
            rope_masked = torch.gather(rope, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, rope_dim))
        else:
            rope_masked = self.rope.pos_embed

        # Append class and register tokens
        if self.class_token is not None:
            if self.pos_embed_special_tokens is True:
                cls_token = self.class_token + pos_embedding[:, self.num_reg_tokens : self.num_reg_tokens + 1, :]
            else:
                cls_token = self.class_token

            batch_class_token = cls_token.expand(x.shape[0], -1, -1)
            x = torch.concat((batch_class_token, x), dim=1)

        if self.reg_tokens is not None:
            if self.pos_embed_special_tokens is True:
                reg_tokens = self.reg_tokens + pos_embedding[:, 0 : self.num_reg_tokens, :]
            else:
                reg_tokens = self.reg_tokens

            batch_reg_tokens = reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        # Apply transformer
        if return_all_features is True:
            xs = self.encoder.forward_features(x, rope_masked)
            xs[-1] = self.norm(xs[-1])
            x = torch.stack(xs, dim=-1)
        else:
            x = self.encoder(x, rope_masked)
            x = self.norm(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            if return_all_features is True:
                x = x[..., -1]

            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        H, W = x.shape[-2:]

        x = self.conv_proj(x)
        x = mask_tensor(x, mask, mask_token=mask_token, patch_factor=self.max_stride // self.stem_stride)

        # Reshape and permute the input tensor
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x, self._get_rope_embed(H, W))
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            features = x[:, self.num_special_tokens :]
            features = features.permute(0, 2, 1)
            B, C, _ = features.size()
            features = features.reshape(B, C, H // self.patch_size, W // self.patch_size)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.attn_pool is not None:
                if self.attn_pool_special_tokens is False:
                    x = x[:, self.num_special_tokens :]

                x = self.attn_pool(x)
                result["embedding"] = x[:, 0]
            elif self.class_token is None:
                x = x[:, self.num_special_tokens :]
                result["embedding"] = x.mean(dim=1)
            else:
                result["embedding"] = x[:, self.num_reg_tokens]

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        H, W = x.shape[-2:]

        # Reshape and permute the input tensor
        x = self.conv_proj(x)
        x = self.patch_embed(x)

        if self.pos_embed_special_tokens is False:
            x = x + self._get_pos_embed(H, W)

        # Expand the class token to the full batch
        if self.class_token is not None:
            batch_class_token = self.class_token.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_class_token, x], dim=1)

        # Expand the register tokens to the full batch
        if self.reg_tokens is not None:
            batch_reg_tokens = self.reg_tokens.expand(x.shape[0], -1, -1)
            x = torch.concat([batch_reg_tokens, x], dim=1)

        if self.pos_embed_special_tokens is True:
            x = x + self._get_pos_embed(H, W)

        x = self.encoder(x, self._get_rope_embed(H, W))
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)

        if self.attn_pool is not None:
            if self.attn_pool_special_tokens is False:
                x = x[:, self.num_special_tokens :]

            x = self.attn_pool(x)
            return x[:, 0]

        if self.class_token is None:
            x = x[:, self.num_special_tokens :]
            return x.mean(dim=1)

        # Classifier "token" as used by standard language architectures
        return x[:, self.num_reg_tokens]

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_special_tokens is True:
            num_prefix_tokens = self.num_special_tokens
        else:
            num_prefix_tokens = 0

        # Add back class tokens
        with torch.no_grad():
            pos_embedding = adjust_position_embedding(
                self.pos_embedding,
                (old_size[0] // self.patch_size, old_size[1] // self.patch_size),
                (new_size[0] // self.patch_size, new_size[1] // self.patch_size),
                num_prefix_tokens,
            )

        self.pos_embedding = nn.Parameter(pos_embedding)

        # Adjust RoPE
        self.rope = RoPE(
            self.hidden_dim // self.num_heads,
            temperature=self.rope_temperature,
            grid_size=(new_size[0] // self.patch_size, new_size[1] // self.patch_size),
            grid_indexing=self.rope_grid_indexing,
            grid_offset=self.rope_grid_offset,
            pt_grid_size=self.pt_grid_size,
            rope_rot_type=self.rope_rot_type,
            device=self.rope.pos_embed.device,
        )

        # Define adjusted decoder block
        self.decoder_block = partial(
            MAEDecoderBlock,
            16,
            num_special_tokens=self.num_special_tokens,
            activation_layer=self.act_layer,
            grid_size=(new_size[0] // self.patch_size, new_size[1] // self.patch_size),
            rope_grid_indexing=self.rope_grid_indexing,
            rope_grid_offset=self.rope_grid_offset,
            rope_temperature=self.rope_temperature,
            layer_scale_init_value=self.layer_scale_init_value,
            norm_layer=self.norm_layer,
            norm_layer_eps=self.norm_layer_eps,
            mlp_layer=self.mlp_layer,
            rope_rot_type=self.rope_rot_type,
        )


# Register model configs (side effects)
register_rope_vit_configs(RoPE_ViT)

registry.register_weights(
    "rope_vit_reg4_b14_capi",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi/resolve/main",
        "description": (
            "RoPE ViT b14 image encoder pre-trained using CAPI. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 327.0,
                "sha256": "175378d81734649567bfe82aac9557f9b0bf48dbd562f26e338b1958fa057472",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-places365",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-places365/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the Places365 dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 328.1,
                "sha256": "4d3ef700eb0d454c9406e9b5c11f70106b46ed4a6ca24c1d89a60097ad78ea9a",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-places365"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-inat21-224px",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-inat21/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the iNaturalist 2021 dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 357.2,
                "sha256": "fb98a4f29a1c6e552a4e22eaf614b0f2d2adedefe2d510fa7e69309208dc0f9f",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-inat21-224px"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-inat21",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-inat21/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the iNaturalist 2021 dataset",
        "resolution": (336, 336),
        "formats": {
            "pt": {
                "file_size": 358.2,
                "sha256": "25befb5a460cc80a5a7961db61e747916461bf6967f3d39d9294ee474bd31304",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-inat21"},
    },
)
registry.register_weights(
    "rope_vit_reg4_b14_capi-imagenet21k",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg4_b14_capi-imagenet21k/resolve/main",
        "description": "RoPE ViT b14 model pre-trained using CAPI, then fine-tuned on the ImageNet-21K dataset",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 383.7,
                "sha256": "1325f35f0d8dff3270d6ce645f81865e9b8de7bacf17f94a9f5e2ef0cd66f56d",
            }
        },
        "net": {"network": "rope_vit_reg4_b14", "tag": "capi-imagenet21k"},
    },
)
registry.register_weights(
    "rope_vit_reg8_so150m_p14_swiglu_rms_avg_capi",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg8_so150m_p14_swiglu_rms_avg_capi/resolve/main",
        "description": (
            "RoPE SoViT 150m p14 image encoder pre-trained using CAPI. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 652.5,
                "sha256": "84808bdb7a46c70eb13a67a766c2c3c9a4a9a37a90679e03fd75619aa5517e80",
            }
        },
        "net": {"network": "rope_vit_reg8_so150m_p14_swiglu_rms_avg", "tag": "capi"},
    },
)
registry.register_weights(
    "rope_vit_reg8_so150m_p14_swiglu_rms_ap_rotnet-capi",
    {
        "url": "https://huggingface.co/birder-project/rope_vit_reg8_so150m_p14_swiglu_rms_ap_rotnet-capi/resolve/main",
        "description": (
            "RoPE SoViT 150m p14 image encoder pre-trained using CAPI, then trained to estimate image orientation"
        ),
        "resolution": (252, 252),
        "formats": {
            "pt": {
                "file_size": 680.9,
                "sha256": "57465120826faa1e61accfb0e51b529c6ae431cc1f6960e4cdd5278d8dbd1edf",
            }
        },
        "net": {"network": "rope_vit_reg8_so150m_p14_swiglu_rms_ap", "tag": "rotnet-capi"},
    },
)

# Perception Encoder: The best visual embeddings are not at the output of the network, by Meta FAIR
# https://arxiv.org/abs/2504.13181
registry.register_weights(
    "rope_i_vit_s16_pn_aps_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_s16_pn_aps_c1_pe-core/resolve/main",
        "description": (
            "ViT s16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 90.0,
                "sha256": "e4429b0bafb9f827698dde73c882c70deb994329ea0dd169f68e76ad256bbb74",
            },
        },
        "net": {"network": "rope_i_vit_s16_pn_aps_c1", "tag": "pe-core"},
    },
)
registry.register_weights(
    "rope_i_vit_reg1_s16_pn_npn_avg_c1_pe-spatial",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_reg1_s16_pn_npn_avg_c1_pe-spatial/resolve/main",
        "description": (
            "ViT s16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (512, 512),
        "formats": {
            "pt": {
                "file_size": 83.9,
                "sha256": "4e65e500f2a7d2b11fc28aaa0b1ad4921692780507de014ebc5659e757327fde",
            },
        },
        "net": {"network": "rope_i_vit_reg1_s16_pn_npn_avg_c1", "tag": "pe-spatial"},
    },
)
registry.register_weights(
    "rope_i_vit_b16_pn_aps_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_b16_pn_aps_c1_pe-core/resolve/main",
        "description": (
            "ViT b16 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 354.4,
                "sha256": "d1c1ba1e8c841f495ff3c0e5e6963a39c8d1ae07dea30d3b82422017a4062d97",
            },
        },
        "net": {"network": "rope_i_vit_b16_pn_aps_c1", "tag": "pe-core"},
    },
)
registry.register_weights(
    "rope_i_vit_l14_pn_aps_c1_pe-core",
    {
        "url": "https://huggingface.co/birder-project/rope_i_vit_l14_pn_aps_c1_pe-core/resolve/main",
        "description": (
            "ViT l14 image encoder pre-trained by Meta FAIR using CLIP. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (336, 336),
        "formats": {
            "pt": {
                "file_size": 1206.0,
                "sha256": "26c2188116cb254d2870c23cc3ab7d60d9ee0606c803b8dbe359e5716498b5c4",
            },
        },
        "net": {"network": "rope_i_vit_l14_pn_aps_c1", "tag": "pe-core"},
    },
)



"""
BIOSCAN-5M benchmark using AMI clustering for unsupervised embedding evaluation

Paper "BIOSCAN-5M: A Multimodal Dataset for Insect Biodiversity",
https://arxiv.org/abs/2406.12723
"""

import argparse
import logging
import time
from pathlib import Path
from typing import Any

import numpy as np
import polars as pl
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.data.datasets.directory import class_to_idx_from_paths
from birder.data.datasets.directory import make_image_dataset
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.ami import evaluate_ami

logger = logging.getLogger(__name__)


def _print_summary_table(results: list[dict[str, Any]]) -> None:
    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("BIOSCAN-5M (AMI)", style="dim")
    table.add_column("AMI Score", justify="right")
    table.add_column("Classes", justify="right")
    table.add_column("Samples", justify="right")

    for result in results:
        table.add_row(
            Path(result["embeddings_file"]).name,
            f"{result['ami_score']:.4f}",
            str(result["num_classes"]),
            str(result["num_samples"]),
        )

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], output_path: Path) -> None:
    rows: list[dict[str, Any]] = []
    for result in results:
        rows.append(
            {
                "embeddings_file": result["embeddings_file"],
                "method": result["method"],
                "ami_score": result["ami_score"],
                "l2_normalize": result["l2_normalize"],
                "num_classes": result["num_classes"],
                "num_samples": result["num_samples"],
            }
        )

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _load_bioscan5m_metadata(data_path: str) -> pl.DataFrame:
    """
    Load metadata from an ImageFolder-compatible directory

    Returns DataFrame with columns: id (filename stem), label
    """

    class_to_idx = class_to_idx_from_paths([data_path])
    image_dataset = make_image_dataset([data_path], class_to_idx)

    rows: list[dict[str, Any]] = []
    for i in range(len(image_dataset)):
        path = image_dataset.paths[i].decode("utf-8")
        label = image_dataset.labels[i].item()
        rows.append({"id": Path(path).stem, "label": label})

    return pl.DataFrame(rows)


def _load_embeddings_with_labels(embeddings_path: str, metadata_df: pl.DataFrame) -> tuple[np.ndarray, np.ndarray, int]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    joined = metadata_df.join(emb_df, on="id", how="inner")
    if joined.height < metadata_df.height:
        logger.warning(f"Join dropped {metadata_df.height - joined.height} samples (missing embeddings)")

    features = joined.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    labels = joined.get_column("label").to_numpy().astype(np.int_)

    num_classes = len(metadata_df.get_column("label").unique())
    logger.info(f"Loaded {features.shape[0]} samples with {features.shape[1]} dimensions, {num_classes} classes")

    return (features, labels, num_classes)


def evaluate_bioscan5m(args: argparse.Namespace) -> None:
    tic = time.time()

    logger.info(f"Loading dataset from {args.data_path}")
    metadata_df = _load_bioscan5m_metadata(args.data_path)
    logger.info(f"Loaded metadata for {metadata_df.height} images")

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        features, labels, num_classes = _load_embeddings_with_labels(embeddings_path, metadata_df)

        logger.info(
            f"Evaluating AMI with umap_dim={args.umap_dim}, seed={args.seed}, "
            f"l2_normalize={not args.no_l2_normalize}"
        )
        ami_score = evaluate_ami(
            features,
            labels,
            n_clusters=num_classes,
            umap_dim=args.umap_dim,
            l2_normalize_features=not args.no_l2_normalize,
            seed=args.seed,
        )
        logger.info(f"AMI score: {ami_score:.4f}")

        results.append(
            {
                "embeddings_file": str(embeddings_path),
                "method": "ami",
                "ami_score": ami_score,
                "l2_normalize": not args.no_l2_normalize,
                "num_classes": num_classes,
                "num_samples": len(labels),
            }
        )

    _print_summary_table(results)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("bioscan5m.csv")
        _write_results_csv(results, output_path)

    toc = time.time()
    logger.info(f"BIOSCAN-5M benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "bioscan5m",
        allow_abbrev=False,
        help="run BIOSCAN-5M benchmark - unsupervised embedding evaluation using AMI clustering",
        description="run BIOSCAN-5M benchmark - unsupervised embedding evaluation using AMI clustering",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval bioscan5m --embeddings "
            "results/embeddings.parquet --data-path ~/Datasets/BIOSCAN-5M/species/testing_unseen --dry-run\n"
            "python -m birder.eval bioscan5m --embeddings results/bioscan5m/*.parquet "
            "--data-path ~/Datasets/BIOSCAN-5M/species/testing_unseen --seed 0\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings parquet files"
    )
    subparser.add_argument("--data-path", type=str, metavar="PATH", help="path to ImageFolder-compatible directory")
    subparser.add_argument("--umap-dim", type=int, default=50, help="target dimensionality for UMAP reduction")
    subparser.add_argument(
        "--no-l2-normalize",
        default=False,
        action="store_true",
        help="disable L2 normalization of embeddings before UMAP",
    )
    subparser.add_argument("--seed", type=int, help="random seed for UMAP")
    subparser.add_argument(
        "--dir", type=str, default="bioscan5m", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.data_path is None:
        raise cli.ValidationError("--data-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_bioscan5m(args)
